import re
import importlib
from pathlib import Path
from hashlib import sha1

def generate_extensions():
    Path('_includes').mkdir(parents=True, exist_ok=True)

    with Path('_includes/ext_toc.rst').open('wt') as f_ext:
        f_ext.write('''
.. toctree::
   :maxdepth: 1
   :caption: Extensions

    ''')
        for line in open('extensions.txt'):
            if '#' in line:
                pkg, display_name = line.split('#', 1)
            else:
                pkg, display_name = line, line
            pkg = pkg.strip()
            display_name = display_name.strip()
            if '<' in display_name:
                display_name, url = display_name.split('<')
                display_name = display_name.strip()
                url = url.strip('>')
            else:
                url = None

            if pkg == '':
                if not url:
                    print(f'Skipping {line!r} -- must be in the format of "# Package Name <URL>"')
                else:
                    f_ext.write(f'   {display_name} <{url}>\n')
                continue

            try:
                metadata = importlib.metadata.metadata(pkg)
                docs = importlib.resources.files(pkg).joinpath('pt_docs')
            except (importlib.metadata.PackageNotFoundError, ModuleNotFoundError):
                if not url:
                    print(f'Skipping {line!r} -- package not installed and no fallback URL provided (format: "# package_name <url>"). You may want to run pip -r extensions.txt for a complete documentation build.')
                else:
                    print(f'Falling back on provided url for {pkg} since the package is not installed. You may want to run pip -r extensions.txt for a complete documentation build.')
                    f_ext.write(f'   {display_name} <{url}>\n')
                continue

            pkg_name = metadata['name']
            if docs.is_dir():
                # Documentation included in the package, copy it over
                paths = [(docs, Path(f'ext/{pkg_name}'))]
                while paths:
                    src, dest = paths.pop()
                    dest.mkdir(parents=True, exist_ok=True)
                    for path in src.iterdir():
                        if path.is_dir():
                            paths.append((path, dest/path.name))
                        else:
                            if (dest/path.name).exists():
                                source_hash = sha1(path.open('rb').read()).hexdigest()
                                dest_hash = sha1((dest/path.name).open('rb').read()).hexdigest()
                                if source_hash == dest_hash:
                                    continue
                            with path.open('rb') as fin, (dest/path.name).open('wb') as fout:
                                fout.write(fin.read())
                f_ext.write(f'   {display_name} <../ext/{pkg_name}/index.rst>\n')
            elif 'home-page' in metadata:
                # No documentation included in the package, but we can link to the repo
                f_ext.write(f'   {display_name} <{metadata["home-page"]}>\n')
            elif url:
                # No documentation included in the package and no home-page, but still can link to the provided <url>> from the config
                f_ext.write(f'   {display_name} <{url}>\n')
            else:
                print(f'Skipping {line!r} -- No pt_docs in package and no home-page or project-url in metadata"')



# java stuff
from pyterrier.terrier import java
from pyterrier.terrier._index import TerrierIndex, TerrierModel
from pyterrier.terrier._text_loader import TerrierTextLoader, terrier_text_loader
from pyterrier.terrier.java import configure, set_version, set_helper_version, extend_classpath, J, set_property, set_properties, run, version, check_version, check_helper_version
from pyterrier.terrier.retriever import Retriever, FeaturesRetriever, TextScorer
from pyterrier.terrier.index_factory import IndexFactory
from pyterrier.terrier.stemmer import TerrierStemmer
from pyterrier.terrier.tokeniser import TerrierTokeniser
from pyterrier.terrier.stopwords import TerrierStopwords
from pyterrier.terrier import index
from pyterrier.terrier.index import TerrierIndexer, FilesIndexer, TRECCollectionIndexer, DFIndexer, DFIndexUtils, IterDictIndexer, IndexingType, treccollection2textgen
from pyterrier.terrier import rewrite
from deprecated import deprecated
import pyterrier as pt


@deprecated(version='0.11.0', reason="use pt.terrier.Retriever() instead")
class TerrierRetrieve(Retriever):
    
    @staticmethod
    @deprecated(version='0.11.0', reason="use pt.terrier.Retriever.from_dataset() instead")
    def from_dataset(*args, **kwargs):
        return Retriever.from_dataset(*args, **kwargs)


@deprecated(version='0.11.0', reason="use pt.terrier.Retriever() instead")
class BatchRetrieve(Retriever):

    @staticmethod
    @deprecated(version='0.11.0', reason="use pt.terrier.Retriever.from_dataset() instead")
    def from_dataset(*args, **kwargs):
        return Retriever.from_dataset(*args, **kwargs)


@deprecated(version='0.11.0', reason="use pt.terrier.FeaturesRetriever() instead")
class FeaturesBatchRetrieve(FeaturesRetriever):
    
    @staticmethod
    @deprecated(version='0.11.0', reason="use pt.terrier.FeaturesRetriever.from_dataset() instead")
    def from_dataset(*args, **kwargs):
        return FeaturesRetriever.from_dataset(*args, **kwargs)


@deprecated(version='0.12.0', reason="This class provides no functionality; inherit from pt.Transformer and set a verbose flag in your constructor instead")
class RetrieverBase(pt.Transformer):
    def __init__(self, verbose=0, **kwargs):
        super().__init__(kwargs)
        self.verbose = verbose


@deprecated(version='0.12.0', reason="This class provides no functionality; inherit from pt.Transformer and set a verbose flag in your constructor instead")
class BatchRetrieveBase(pt.Transformer):
    def __init__(self, verbose=0, **kwargs):
        super().__init__(kwargs)
        self.verbose = verbose


__all__ = [
    # java stuff
    'java', 'configure', 'set_version', 'set_helper_version', 'extend_classpath', 'J', 'version', 'check_version', 'check_helper_version',

    # High-level API
    'TerrierIndex', 'TerrierModel', 'TerrierStemmer', 'TerrierStopwords', 'TerrierTokeniser',

    # retrieval
    'BatchRetrieveBase', 'Retriever', 'RetrieverBase', 'BatchRetrieve', 'TerrierRetrieve', 'FeaturesRetriever', 'FeaturesBatchRetrieve', 'TextScorer',

    # indexing
    'index', 'TerrierIndexer', 'FilesIndexer', 'TRECCollectionIndexer', 'DFIndexer', 'DFIndexUtils', 'IterDictIndexer', 'IndexingType', 'treccollection2textgen',

    # rewriting
    'rewrite',

    # misc
    'IndexFactory', 'set_property', 'set_properties', 'run', 
    'TerrierTextLoader', 'terrier_text_loader',
]



import pyterrier as pt
import tempfile
import unittest
from .base import TempDirTestCase
import os
import pandas as pd
import shutil
import tempfile


class TestIndexOp(TempDirTestCase):

    def test_index_corpus_iter(self):
        import sys
        MIN_PYTHON = (3, 8)
        if sys.version_info < MIN_PYTHON:
            self.skipTest("Not minimum Python requirements")

        documents = [
            {'docno' : 'd1', 'text': 'stemming stopwords stopwords'},
        ]
        index = pt.IndexFactory.of( pt.IterDictIndexer(tempfile.mkdtemp(), stopwords=None, stemmer=None).index(documents) )
        self.assertEqual(1, len(index))
        self.assertEqual(2, index.getCollectionStatistics().getNumberOfUniqueTerms())
        self.assertEqual(3, index.getCollectionStatistics().getNumberOfTokens())

        # check that get_corpus_iter() contains the correct information 
        iter = index.get_corpus_iter()
        first_doc = next(iter)
        self.assertTrue(first_doc is not None)
        self.assertIn('docno', first_doc)
        self.assertIn('toks', first_doc)
        self.assertIn('stemming', first_doc['toks'])
        self.assertIn('stopwords', first_doc['toks'])
        self.assertEqual(1, first_doc['toks']['stemming'])
        self.assertEqual(2, first_doc['toks']['stopwords'])
        with(self.assertRaises(StopIteration)):
            next(iter)

        #  now check that a static pruning pipe can operate as expected. this example comes from terrier-index-api.rst
        index_pipe = (
            # update the toks column for each document, keeping only terms with frequency > 1
            pt.apply.toks(lambda row: { t : row['toks'][t] for t in row['toks'] if row['toks'][t] > 1 } ) 
            >> pt.IterDictIndexer(tempfile.mkdtemp(), pretokenised=True)
        )
        new_index_ref = index_pipe.index( index.get_corpus_iter())
        pruned_index = pt.IndexFactory.of(new_index_ref)
        self.assertEqual(1, len(pruned_index))
        self.assertEqual(1, pruned_index.getCollectionStatistics().getNumberOfUniqueTerms())
        self.assertEqual(2, pruned_index.getCollectionStatistics().getNumberOfTokens())

    def test_index_corpus_iter_empty(self):
        import sys
        MIN_PYTHON = (3, 8)
        if sys.version_info < MIN_PYTHON:
            self.skipTest("Not minimum Python requirements")
            
        # compared to test_index_corpus_iter, this tests empty documents are handled correctly.
        documents = [
            {'docno' : 'd0', 'text':''},
            {'docno' : 'd1', 'text':''},
            {'docno' : 'd2', 'text': 'stemming stopwords stopwords'},
            {'docno' : 'd3', 'text':''},
            {'docno' : 'd4', 'text': 'stemming stopwords stopwords'},
            {'docno' : 'd5', 'text': ''}
        ]
        index = pt.IndexFactory.of( pt.IterDictIndexer(tempfile.mkdtemp(), stopwords=None, stemmer=None).index(documents) )
        self.assertEqual(6, len(index))
        self.assertEqual(2, index.getCollectionStatistics().getNumberOfUniqueTerms())
        self.assertEqual(6, index.getCollectionStatistics().getNumberOfTokens())

        iter = index.get_corpus_iter()

        counter = 0
        for doc in documents:
            next_doc = next(iter)
            counter += 1
            self.assertTrue(next_doc is not None)
            self.assertIn('docno', next_doc)
            self.assertIn('toks', next_doc)
            if doc['text'] == '':
                self.assertEqual(0, len(next_doc['toks']))
            else:
                self.assertIn('stemming', next_doc['toks'])
                self.assertIn('stopwords', next_doc['toks'])
                self.assertEqual(1, next_doc['toks']['stemming'])
                self.assertEqual(2, next_doc['toks']['stopwords'])

        with(self.assertRaises(StopIteration)):
            next(iter)
        self.assertEqual(counter, len(documents))

        #  now check that a static pruning pipe can operate as expected. this example comes from terrier-index-api.rst
        index_pipe = (
            # update the toks column for each document, keeping only terms with frequency > 1
            pt.apply.toks(lambda row: { t : row['toks'][t] for t in row['toks'] if row['toks'][t] > 1 } ) 
            >> pt.IterDictIndexer(tempfile.mkdtemp(), pretokenised=True)
        )
        new_index_ref = index_pipe.index( index.get_corpus_iter())
        pruned_index = pt.IndexFactory.of(new_index_ref)
        self.assertEqual(6, len(pruned_index))
        self.assertEqual(1, pruned_index.getCollectionStatistics().getNumberOfUniqueTerms())
        self.assertEqual(4, pruned_index.getCollectionStatistics().getNumberOfTokens())

    def test_index_add_write(self):
        # inspired by https://github.com/terrier-org/pyterrier/issues/390
        documents = [
            {'text': 'Creates a Function that returns a Function or returns the value of the given property .'},
            {'text': 'Returns the URL of the occupants .'},
            {'text': 'Exit the timer with the default values .'},
        ]

        # Create new index from pandas dataframe
        pd_indexer = pt.IterDictIndexer(tempfile.mkdtemp(), blocks=True)
        df = pd.DataFrame(documents)
        df['docno'] = df.index.astype(str)
        indexref = pd_indexer.index(df.to_dict(orient='records'))
        index1 = pt.IndexFactory.of(indexref)

        # Create new index from pandas dataframe
        pd_indexer = pt.IterDictIndexer(tempfile.mkdtemp(), blocks=True)
        df = pd.DataFrame(documents)
        df['docno'] = df.index.astype(str)
        indexref = pd_indexer.index(df.to_dict(orient='records'))
        index2 = pt.IndexFactory.of(indexref)

        # Merge indexes
        comb_index = index1 + index2

        new_disk_index_loc = tempfile.mkdtemp()
        self.assertEqual(len(index1) + len(index2), len(comb_index))
        
        # Instantiate writer object and write merged index to disk
        writer = pt.java.autoclass("org.terrier.structures.indexing.DiskIndexWriter")(new_disk_index_loc, "data")
        writer.write(comb_index)

        new_disk_index = pt.IndexFactory.of(new_disk_index_loc)

        self.assertEqual(len(index1) + len(index2), len(new_disk_index))


from enum import Enum
import pyterrier as pt


_stemmer_cache = {}


class TerrierStemmer(Enum):
    """A built-in Terrier stemmer.

    The stemming configuration is saved in the index and loaded at retrieval time. `Snowball <https://snowballstem.org/>`_
    stemmers for various languages  `are available in Terrier <http://terrier.org/docs/current/javadoc/org/terrier/terms/package-summary.html>`_.

    This enum is primarily used with :class:`~pyterrier.terrier.TerrierIndexer.indexer`.
    """
    none = 'none' #: Apply no stemming
    porter = 'porter' #: Apply Porter's English stemmer
    weakporter = 'weakporter' #: Apply a weak version of Porter's English stemmer
    # available snowball stemmers in Terrier
    danish = 'danish' #: Snowball Danish stemmer
    finnish = 'finnish' #: Snowball Finnish stemmer
    german = 'german' #: Snowball German stemmer
    hungarian = 'hungarian' #: Snowball Hungarian stemmer
    norwegian = 'norwegian' #: Snowball Norwegian stemmer
    portugese = 'portugese' #: Snowball Portuguese stemmer
    spanish = 'spanish' #: Snowball Spanish stemmer
    swedish = 'swedish' #: Snowball Swedish stemmer
    turkish = 'turkish' #: Snowball Turkish stemmer

    @staticmethod
    def _to_obj(this):
        try:
            return TerrierStemmer(this)
        except ValueError:
            return this

    @staticmethod
    def _to_class(this):
        if this is None or this == TerrierStemmer.none:
            return None
        if this == TerrierStemmer.porter:
            return 'PorterStemmer'
        if this == TerrierStemmer.weakporter:
            return 'WeakPorterStemmer'
        
        # snowball stemmers
        if this == TerrierStemmer.danish:
            return 'DanishSnowballStemmer'
        if this == TerrierStemmer.finnish:
            return 'FinnishSnowballStemmer'
        if this == TerrierStemmer.german:
            return 'GermanSnowballStemmer'
        if this == TerrierStemmer.hungarian:
            return 'HungarianSnowballStemmer'
        if this == TerrierStemmer.norwegian:
            return 'NorwegianSnowballStemmer'
        if this == TerrierStemmer.portugese:
            return 'PortugueseSnowballStemmer'
        if this == TerrierStemmer.spanish:
            return 'SpanishSnowballStemmer'
        if this == TerrierStemmer.swedish:
            return 'SwedishSnowballStemmer'
        if this == TerrierStemmer.turkish:
            return 'TurkishSnowballStemmer'

        if isinstance(this, str):
            return this

    @pt.java.required
    def stem(self, tok):
        """Stem a single token using this stemmer.

        .. code-block:: python

            stemmer = pt.TerrierStemmer.porter
            stemmed_word = stemmer.stem('abandoned')
        """
        if self not in _stemmer_cache:
            clz_name = self._to_class(self)
            if clz_name is None:
                _stemmer_cache[self] = _NoOpStem()
            else:
                if '.' not in clz_name:
                    clz_name = f'org.terrier.terms.{clz_name}'
                 # stemmers are termpipeline objects, and these have chained constructors
                 # pass None to use the appropriate constructor
                _stemmer_cache[self] = pt.java.autoclass(clz_name)(None)
        return _stemmer_cache[self].stem(tok)


class _NoOpStem():
    def stem(self, word):
        return word



import os
import json
from packaging.version import Version
from typing import Optional, Union, List, Dict, Any
import pyterrier as pt

TERRIER_PKG = "org.terrier"

_SAVED_FNS = []

_properties = None

configure = pt.java.register_config('pt.terrier.java', {
    'terrier_version': os.environ.get("TERRIER_VERSION") or None,
    'helper_version': os.environ.get("TERRIER_HELPER_VERSION") or None,
    'boot_packages': [],
    'force_download': True,
})

@pt.java.before_init
def set_version(version: Optional[str] = None):
    configure['terrier_version'] = version


@pt.java.before_init
def set_helper_version(version: Optional[str] = None):
    configure['helper_version'] = version


class TerrierJavaInit(pt.java.JavaInitializer):
    def priority(self) -> int:
        return -10 # between pt.java.core (-100) and default (0) to load earlier than extensions

    def pre_init(self, jnius_config):

        # If version is not specified, find newest and download it
        if configure['terrier_version'] is None:
            terrier_version = pt.java.mavenresolver.latest_version_num(TERRIER_PKG, "terrier-assemblies")
        else:
            terrier_version = str(configure['terrier_version']) # just in case its a float
        configure['terrier_version'] = terrier_version # save this specific version

        # obtain the fat jar from Maven
        # "snapshot" means use Jitpack.io to get a build of the current
        # 5.x branch from Github - see https://jitpack.io/#terrier-org/terrier-core/5.x-SNAPSHOT
        if terrier_version == "snapshot":
            trJar = pt.java.mavenresolver.get_package_jar("com.github.terrier-org.terrier-core", "terrier-assemblies", "5.x-SNAPSHOT", artifact="jar-with-dependencies", force_download=configure['force_download'])
        else:
            trJar = pt.java.mavenresolver.get_package_jar(TERRIER_PKG, "terrier-assemblies", terrier_version, artifact="jar-with-dependencies")
        jnius_config.add_classpath(trJar)

        # now the helper classes
        if configure['helper_version'] is None or configure['helper_version'] == 'snapshot':
            helper_version = pt.java.mavenresolver.latest_version_num(TERRIER_PKG, "terrier-python-helper")
            configure['helper_version'] = helper_version # save this specific version
        else:
            helper_version = str(configure['helper_version']) # just in case its a float
        helper_jar = pt.java.mavenresolver.get_package_jar(TERRIER_PKG, 'terrier-python-helper', helper_version)
        jnius_config.add_classpath(helper_jar)

        # This is for parallel -- it means that when re-configured in a parallel process, force_download will be False
        # and mavenresolver will use the version that was just downloaded above (not try to do it again).
        configure['force_download'] = False


    @pt.java.required_raise
    def post_init(self, jnius):
        global _properties

        jnius.protocol_map["org.terrier.structures.postings.IterablePosting"] = {
            '__iter__': lambda self: self,
            '__next__': lambda self: _iterableposting_next(self),
            '__str__': lambda self: self.toString()
        }

        jnius.protocol_map["org.terrier.structures.CollectionStatistics"] = {
            '__str__': lambda self: self.toString()
        }

        jnius.protocol_map["org.terrier.structures.LexiconEntry"] = {
            '__str__': lambda self: self.toString()
        }

        jnius.protocol_map["org.terrier.structures.Lexicon"] = {
            '__getitem__': _lexicon_getitem,
            '__contains__': lambda self, term: self.getLexiconEntry(term) is not None,
            '__len__': lambda self: self.numberOfEntries()
        }

        jnius.protocol_map["org.terrier.querying.IndexRef"] = {
            '__eq__' : lambda self, other: self.equals(other),
            '__reduce__' : _index_ref_reduce,
            '__getstate__' : lambda self : None,
            'text_loader': pt.terrier.terrier_text_loader,
        }

        jnius.protocol_map["org.terrier.matching.models.WeightingModel"] = {
            '__reduce__' : _wmodel_reduce,
            '__getstate__' : lambda self : None,
        }

        jnius.protocol_map["org.terrier.python.CallableWeightingModel"] = {
            '__reduce__' : _callable_wmodel_reduce,
            '__getstate__' : lambda self : None,
        }

        jnius.protocol_map["org.terrier.structures.Index"] = {
            # this means that len(index) returns the number of documents in the index
            '__len__': lambda self: self.getCollectionStatistics().getNumberOfDocuments(),

            # document-wise composition of indices: adding more documents to an index, by merging two indices with 
            # different numbers of documents. This implemented by the overloading the `+` Python operator
            '__add__': _index_add,

            # get_corpus_iter returns a yield generator that return {"docno": "d1", "toks" : {'a' : 1}}
            'get_corpus_iter' : _index_corpusiter,
            'text_loader': pt.terrier.terrier_text_loader,
        }

        self._post_init_index(jnius)

        pt.IndexRef = J.IndexRef
        _properties = pt.java.J.Properties()
        pt.ApplicationSetup = J.ApplicationSetup
        J.ApplicationSetup.bootstrapInitialisation(_properties)

    @pt.java.required_raise
    def message(self):
        version_string = J.Version.VERSION
        if "BUILD_DATE" in dir(J.Version):
            version_string += f" (build: {J.Version.BUILD_USER} {J.Version.BUILD_DATE})"

        return f"version={version_string}, helper_version={configure['helper_version']}"

    def _post_init_index(self, jnius):
        @pt.java.required
        class DocListIterator(jnius.PythonJavaClass):
            __javainterfaces__ = [
                'java/util/Iterator',
            ]

            def __init__(self, pyiterator):
                self.pyiterator = pyiterator
                self.hasnext = True
                self.lastdoc = None
                self.tr57 = not pt.terrier.check_version("5.8")

            @staticmethod 
            def pyDictToMap(a_dict): #returns Map<String,String>
                rtr = pt.java.J.HashMap()
                for k,v in a_dict.items():
                    rtr.put(k, v)
                return rtr

            def pyDictToMapEntry(self,doc_dict : Dict[str,Any]): #returns Map.Entry<Map<String,String>, DocumentPostingList>>
                dpl = pt.terrier.J.DocumentPostingList()
                # this works around a bug in the counting of doc lengths in Tr 5.7
                if self.tr57:
                    for t, tf in doc_dict["toks"].items():
                        for i in range(int(tf)):
                            dpl.insert(t)
                else: # this code for 5.8 onwards
                    for t, tf in doc_dict["toks"].items():
                        dpl.insert(int(tf), t)
                
                # we cant make the toks column into the metaindex as it isnt a string. remove it.
                del doc_dict["toks"]
                return pt.terrier.J.MapEntry(DocListIterator.pyDictToMap(doc_dict), dpl)

            @jnius.java_method('()Z')
            def hasNext(self):
                return self.hasnext

            @jnius.java_method('()Ljava/lang/Object;')
            def next(self):
                try:
                    doc_dict = next(self.pyiterator)
                except StopIteration:
                    self.hasnext = False
                    # terrier will ignore a null return from an iterator
                    return None
                # keep this around to prevent being GCd before Java can read it
                self.lastdoc = self.pyDictToMapEntry(doc_dict)
                return self.lastdoc
                
        @pt.java.required
        class TQDMCollection(jnius.PythonJavaClass):
            __javainterfaces__ = ['org/terrier/indexing/Collection']

            def __init__(self, collection):
                super(TQDMCollection, self).__init__()
                assert isinstance(collection, pt.terrier.J.MultiDocumentFileCollection)
                self.collection = collection
                size = self.collection.FilesToProcess.size()
                self.pbar = pt.tqdm(total=size, unit="files")
                self.last = -1
            
            @jnius.java_method('()Z')
            def nextDocument(self):
                rtr = self.collection.nextDocument()
                filenum = self.collection.FileNumber
                if filenum > self.last:
                    self.pbar.update(filenum - self.last)
                    self.last = filenum
                return rtr

            @jnius.java_method('()V')
            def reset(self):
                self.pbar.reset()
                self.collection.reset()

            @jnius.java_method('()V')
            def close(self):
                self.pbar.close()
                self.collection.close()

            @jnius.java_method('()Z')
            def endOfCollection(self):
                return self.collection.endOfCollection()

            @jnius.java_method('()Lorg/terrier/indexing/Document;')
            def getDocument(self):
                global lastdoc
                lastdoc = self.collection.getDocument()
                return lastdoc

        class PythonListIterator(jnius.PythonJavaClass):
            __javainterfaces__ = ['java/util/Iterator']

            def __init__(self, text, meta, convertFn, len=None, index=0):
                super(PythonListIterator, self).__init__()
                self.text = text
                self.meta = meta
                self.index = index
                self.convertFn = convertFn
                if len is None:
                    self.len = len(self.text)
                else:
                    self.len = len

            @jnius.java_method('()V')
            def remove():
                # 1
                pass

            @jnius.java_method('(Ljava/util/function/Consumer;)V')
            def forEachRemaining(action):
                # 1
                pass

            @jnius.java_method('()Z')
            def hasNext(self):
                return self.index < self.len

            @jnius.java_method('()Ljava/lang/Object;')
            def next(self):
                text = self.text[self.index]
                meta = self.meta.__next__()
                self.index += 1
                global lastdoc
                if self.convertFn is not None:
                    lastdoc = self.convertFn(text, meta)
                else:
                    lastdoc = [text, meta]
                return lastdoc

        @pt.java.required
        class FlatJSONDocumentIterator(jnius.PythonJavaClass):
            __javainterfaces__ = ['java/util/Iterator']

            def __init__(self, it):
                super(FlatJSONDocumentIterator, self).__init__()
                self._it = it
                # easiest way to support hasNext is just to start consuming right away, I think
                self._next = next(self._it, StopIteration)

            @jnius.java_method('()V')
            def remove():
                # 1
                pass

            @jnius.java_method('(Ljava/util/function/Consumer;)V')
            def forEachRemaining(action):
                # 1
                pass

            @jnius.java_method('()Z')
            def hasNext(self):
                return self._next is not StopIteration

            @jnius.java_method('()Ljava/lang/Object;')
            def next(self):
                result = self._next
                self._next = next(self._it, StopIteration)
                if result is not StopIteration:
                    global lastdoc
                    lastdoc = pt.terrier.J.FlatJSONDocument(json.dumps(result))
                    return lastdoc
                return None

        class TQDMSizeCollection(jnius.PythonJavaClass):
            __javainterfaces__ = ['org/terrier/indexing/Collection']

            def __init__(self, collection, total):
                super(TQDMSizeCollection, self).__init__()
                self.collection = collection
                self.pbar = pt.tqdm(total=total, unit="documents")
            
            @jnius.java_method('()Z')
            def nextDocument(self):
                rtr = self.collection.nextDocument()
                self.pbar.update()
                return rtr

            @jnius.java_method('()V')
            def reset(self):
                self.pbar.reset()
                self.collection.reset()

            @jnius.java_method('()V')
            def close(self):
                self.pbar.close()
                self.collection.close()

            @jnius.java_method('()Z')
            def endOfCollection(self):
                return self.collection.endOfCollection()

            @jnius.java_method('()Lorg/terrier/indexing/Document;')
            def getDocument(self):
                global lastdoc
                lastdoc = self.collection.getDocument()
                return lastdoc

        pt.terrier.index.DocListIterator = DocListIterator
        pt.terrier.index.PythonListIterator = PythonListIterator
        pt.terrier.index.FlatJSONDocumentIterator = FlatJSONDocumentIterator
        pt.terrier.index.TQDMCollection = TQDMCollection
        pt.terrier.index.TQDMSizeCollection = TQDMSizeCollection


def _new_indexref(s):
    return pt.IndexRef.of(s)


@pt.java.required
def _new_wmodel(b):
    return J.Serialization.deserialize(b, J.ApplicationSetup.getClass("org.terrier.matching.models.WeightingModel"))


def _new_callable_wmodel(byterep):
    import dill as pickle
    from dill import extend
    #see https://github.com/SeldonIO/alibi/issues/447#issuecomment-881552005
    extend(use_dill=False)            
    fn = pickle.loads(byterep)
    #we need to prevent these functions from being GCd.
    global _SAVED_FNS
    _SAVED_FNS.append(fn)
    callback, wmodel = pt.terrier.retriever._function2wmodel(fn)
    _SAVED_FNS.append(callback)
    #print("Stored lambda fn  %s and callback in SAVED_FNS, now %d stored" % (str(fn), len(SAVED_FNS)))
    return wmodel


def _iterableposting_next(self):
    ''' dunder method for iterating IterablePosting '''
    nextid = self.next()
    # 2147483647 is IP.EOL. fix this once static fields can be read from instances.
    if 2147483647 == nextid:
        raise StopIteration()
    return self


def _lexicon_getitem(self, term):
    ''' dunder method for accessing Lexicon '''
    rtr = self.getLexiconEntry(term)
    if rtr is None:
        raise KeyError()
    return rtr


def _index_ref_reduce(self):
    return (
        _new_indexref,
        (str(self.toString()),),
        None
    )


# handles the pickling of WeightingModel classes, which are themselves usually Serializable in Java
@pt.java.required
def _wmodel_reduce(self):
    serialized = bytes(J.Serialization.serialize(self))
    return (
        _new_wmodel,
        (serialized, ),
        None
    )


def _callable_wmodel_reduce(self):
    # get bytebuffer representation of lambda
    # convert bytebyffer to python bytearray
    serlzd = self.scoringClass.serializeFn()
    bytesrep = pt.java.bytebuffer_to_array(serlzd)
    del(serlzd)
    return (
        _new_callable_wmodel,
        (bytesrep, ),
        None
    )


@pt.java.required
def _index_add(self, other):
    fields_1 = self.getCollectionStatistics().getNumberOfFields()
    fields_2 = self.getCollectionStatistics().getNumberOfFields()
    if fields_1 != fields_2:
        raise ValueError("Cannot document-wise merge indices with different numbers of fields (%d vs %d)" % (fields_1, fields_2))
    blocks_1 = self.getCollectionStatistics().hasPositions()
    blocks_2 = other.getCollectionStatistics().hasPositions()
    if blocks_1 != blocks_2:
        raise ValueError("Cannot document-wise merge indices with and without positions (%r vs %r)" % (blocks_1, blocks_2))
    return J.MultiIndex([self, other], blocks_1, fields_1 > 0)


def _index_corpusiter(self, return_toks=True):
    def _index_corpusiter_meta(self):
        meta_inputstream = self.getIndexStructureInputStream("meta")
        keys = self.getMetaIndex().getKeys()
        keys_offset = { k: offset for offset, k in enumerate(keys) }
        while meta_inputstream.hasNext():
            item = meta_inputstream.next()
            yield {k : item[keys_offset[k]] for k in keys_offset}

    def _index_corpusiter_direct_pretok(self):
        meta_inputstream = self.getIndexStructureInputStream("meta")
        keys = self.getMetaIndex().getKeys()
        keys_offset = { k: offset for offset, k in enumerate(keys) }
        keys_offset = {'docno' : keys_offset['docno'] }
        direct_inputstream = self.getIndexStructureInputStream("direct")
        lex = self.getLexicon()

        ip = None
        while (ip := direct_inputstream.getNextPostings()) is not None: # this is the next() method

            # yield empty toks dicts for empty documents
            for skipped in range(0, direct_inputstream.getEntriesSkipped()):
                meta = meta_inputstream.next()
                rtr = {k : meta[keys_offset[k]] for k in keys_offset}   
                rtr['toks'] = {}
                yield rtr

            toks = {}
            while ip.next() != ip.EOL:
                t, _ = lex[ip.getId()]
                toks[t] = ip.getFrequency()
            meta = meta_inputstream.next()
            rtr = {'toks' : toks}
            rtr.update({k : meta[keys_offset[k]] for k in keys_offset})
            yield rtr

        # yield for trailing empty documents
        for skipped in range(0, direct_inputstream.getEntriesSkipped()):
            meta = meta_inputstream.next()
            rtr = {k : meta[keys_offset[k]] for k in keys_offset}   
            rtr['toks'] = {}
            yield rtr
    
    if return_toks:
        if not self.hasIndexStructureInputStream("direct"):
            raise ValueError("No direct index input stream available, cannot use return_toks=True")
        return _index_corpusiter_direct_pretok(self)
    return _index_corpusiter_meta(self)



@pt.java.required
def extend_classpath(packages: Union[str, List[str]]):
    """
        Allows to add packages to Terrier's classpath after the JVM has started.
    """
    if isinstance(packages, str):
        packages = [packages]
    assert check_version(5.3), "Terrier 5.3 required for this functionality"
    package_list = pt.java.J.ArrayList()
    for package in packages:
        package_list.add(package)
    mvnr = J.ApplicationSetup.getPlugin("MavenResolver")
    assert mvnr is not None
    mvnr = pt.java.cast("org.terrier.utility.MavenResolver", mvnr)
    mvnr.addDependencies(package_list)


@pt.java.required
def set_property(k, v):
    """
        Allows to set a property in Terrier's global properties configuration. Example::

            pt.set_property("termpipelines", "")

        While Terrier has a variety of properties -- as discussed in its 
        `indexing <https://github.com/terrier-org/terrier-core/blob/5.x/doc/configure_indexing.md>`_ 
        and `retrieval <https://github.com/terrier-org/terrier-core/blob/5.x/doc/configure_retrieval.md>`_ 
        configuration guides -- in PyTerrier, we aim to expose Terrier configuration through appropriate 
        methods or arguments. So this method should be seen as a safety-valve - a way to override the 
        Terrier configuration not explicitly supported by PyTerrier.
    """
    _properties[str(k)] = str(v)
    J.ApplicationSetup.bootstrapInitialisation(_properties)


@pt.java.required
def set_properties(kwargs):
    """
        Allows to set many properties in Terrier's global properties configuration
    """
    for key, value in kwargs.items():
        _properties[str(key)] = str(value)
    J.ApplicationSetup.bootstrapInitialisation(_properties)


@pt.java.required
def run(cmd, args=[]):
    """
        Allows to run a Terrier executable class, i.e. one that can be access from the `bin/terrier` commandline programme.
    """
    J.CLITool.main([cmd] + args)


@pt.java.required
def version():
    """
        Returns the version string from the underlying Terrier platform.
    """
    return J.Version.VERSION


def check_version(min):
    """
        Returns True iff the underlying Terrier version is no older than the requested version.
    """
    current_ver = version()
    assert current_ver is not None, "Could not obtain Terrier version"
    current_ver = Version(current_ver.replace("-SNAPSHOT", ""))

    min = Version(str(min))
    return current_ver >= min


def check_helper_version(min):
    """
        Returns True iff the underlying Terrier helper version is no older than the requested version.
    """
    current_ver = configure['helper_version']
    assert current_ver is not None, "Could not obtain Terrier helper version" 
    current_ver = Version(current_ver.replace("-SNAPSHOT", ""))

    min = Version(str(min))
    return current_ver >= min



# Terrier-specific classes
J = pt.java.JavaClasses(
    ApplicationSetup = 'org.terrier.utility.ApplicationSetup',
    IndexRef = 'org.terrier.querying.IndexRef',
    Version = 'org.terrier.Version',
    Tokenizer = 'org.terrier.indexing.tokenisation.Tokeniser',
    Serialization = 'org.terrier.python.Serialization',
    IndexOnDisk = 'org.terrier.structures.IndexOnDisk',
    IndexFactory = 'org.terrier.structures.IndexFactory',
    MultiIndex = 'org.terrier.realtime.multi.MultiIndex',
    CLITool = 'org.terrier.applications.CLITool',
    ApplyTermPipeline = 'org.terrier.querying.ApplyTermPipeline',
    ManagerFactory = 'org.terrier.querying.ManagerFactory',
    Request = 'org.terrier.querying.Request',

    # Indexing
    TaggedDocument = 'org.terrier.indexing.TaggedDocument',
    FlatJSONDocument = 'org.terrier.indexing.FlatJSONDocument',
    Tokeniser = 'org.terrier.indexing.tokenisation.Tokeniser',
    TRECCollection = 'org.terrier.indexing.TRECCollection',
    SimpleFileCollection = 'org.terrier.indexing.SimpleFileCollection',
    BasicIndexer = 'org.terrier.structures.indexing.classical.BasicIndexer',
    BlockIndexer = 'org.terrier.structures.indexing.classical.BlockIndexer',
    BasicSinglePassIndexer = 'org.terrier.structures.indexing.singlepass.BasicSinglePassIndexer',
    BlockSinglePassIndexer = 'org.terrier.structures.indexing.singlepass.BlockSinglePassIndexer',
    BasicMemoryIndexer = lambda: 'org.terrier.realtime.memory.MemoryIndexer' if check_version("5.7") else 'org.terrier.python.MemoryIndexer',
    Collection = 'org.terrier.indexing.Collection',
    StructureMerger = 'org.terrier.structures.merging.StructureMerger',
    BlockStructureMerger = 'org.terrier.structures.merging.BlockStructureMerger',
    DocumentPostingList = 'org.terrier.structures.indexing.DocumentPostingList',
    MapEntry = 'org.terrier.structures.collections.MapEntry',
    MultiDocumentFileCollection = 'org.terrier.indexing.MultiDocumentFileCollection',
    CollectionFactory = 'org.terrier.indexing.CollectionFactory',
    TagSet = 'org.terrier.utility.TagSet',
    CollectionFromDocumentIterator = 'org.terrier.python.CollectionFromDocumentIterator',
    PTUtils = 'org.terrier.python.PTUtils',
    Index = 'org.terrier.structures.Index',
    JsonlDocumentIterator = 'org.terrier.python.JsonlDocumentIterator',
    JsonlPretokenisedIterator = 'org.terrier.python.JsonlPretokenisedIterator',
    ParallelIndexer = 'org.terrier.python.ParallelIndexer',

    # PRF
    TerrierQLParser = 'org.terrier.querying.TerrierQLParser',
    TerrierQLToMatchingQueryTerms = 'org.terrier.querying.TerrierQLToMatchingQueryTerms',
    QueryResultSet = 'org.terrier.matching.QueryResultSet',
    DependenceModelPreProcess = 'org.terrier.querying.DependenceModelPreProcess',
    RM3 = 'org.terrier.querying.RM3',
)


import pyterrier as pt
import tempfile
import unittest
from .base import BaseTestCase
import os
import pandas as pd

class TestIrDatasetsIntegration(BaseTestCase):

    def test_vaswani(self):
        dataset = pt.datasets.get_dataset('irds:vaswani')
        with self.subTest('topics'):
            topics = dataset.get_topics()
            self.assertEqual(len(topics), 93)
        with self.subTest('qrels'):
            qrels = dataset.get_qrels()
            self.assertEqual(len(qrels), 2083)
        with self.subTest('corpus'):
            corpus = dataset.get_corpus_iter()
            self.assertEqual(len(corpus), 11429)
            corpus = list(corpus)
            self.assertEqual(len(corpus), 11429)
            with tempfile.TemporaryDirectory() as d:
                indexer = pt.index.IterDictIndexer(d)
                indexref = indexer.index(dataset.get_corpus_iter())
                index = pt.IndexFactory.of(indexref)
                self.assertEqual(index.lexicon['bit'].frequency, 33)
                index.close()
        with self.subTest('corpus-slice'):
            corpus = dataset.get_corpus_iter(start=1, count=2)
            self.assertEqual(2, len(corpus))
            self.assertEqual(2, len(list(corpus)))
            corpus = dataset.get_corpus_iter(start=0, count=2)
            self.assertEqual(2, len(corpus))
            self.assertEqual(2, len(list(corpus)))

    def test_antique(self):
        dataset = pt.datasets.get_dataset('irds:antique/test')
        with self.subTest('topics - no tokenisation'):
            topics = dataset.get_topics() #tokenise_query=False
            self.assertEqual(len(topics), 200)
            self.assertEqual(topics['query'][0], 'how can we get concentration onsomething?')

    def test_cord19(self):
        if "PYTERRIER_TEST_IRDS_CORD" in os.environ:
            dataset = pt.datasets.get_dataset('irds:cord19/trec-covid')
            with self.subTest('topics'):
                topics = dataset.get_topics()
                self.assertEqual(len(topics), 50)
            with self.subTest('qrels'):
                qrels = dataset.get_qrels()
                self.assertEqual(len(qrels), 69318)
            with self.subTest('corpus'):
                corpus = dataset.get_corpus_iter()
                corpus = list(corpus)
                self.assertEqual(len(corpus), 192509)
            with self.subTest('indexing'):
                with tempfile.TemporaryDirectory() as d:
                    indexer = pt.index.IterDictIndexer(d)
                    indexref = indexer.index(dataset.get_corpus_iter(), fields=('title', 'abstract'))
                    index = pt.IndexFactory.of(indexref)
                    self.assertEqual(index.lexicon['covid'].frequency, 200582)

    def test_results(self):
        if "PYTERRIER_TEST_IRDS_RESULTS" in os.environ:
            dataset = pt.datasets.get_dataset('irds:msmarco-passage/trec-dl-2019/judged')
            with self.subTest('results'):
                results = dataset.get_results()
                self.assertEqual(41042, len(results))
                self.assertEqual(['qid', 'docno', 'score', 'query'], list(results.columns))
                self.assertEqual('1037798', results.iloc[0].qid)
                self.assertEqual('1031599', results.iloc[0].docno)
                self.assertEqual(0., results.iloc[0].score)
                self.assertEqual('who is robert gray', results.iloc[0].query)
                # ensure it's terrier-tokenised (orig text is "tracheids are part of _____.")
                self.assertEqual('tracheids are part of', results[results.qid=='1124210'].iloc[0].query)

    def test_nonexistant(self):
        # Should raise an error when you request an irds: dataset that doesn't exist
        with self.assertRaises(KeyError):
            dataset = pt.datasets.get_dataset('irds:bla-bla-bla')

        # (it's the same erorr raised without using the irds: integration)
        with self.assertRaises(KeyError):
            dataset = pt.datasets.get_dataset('bla-bla-bla')

    def test_variants(self):
        dataset = pt.get_dataset('irds:clueweb09/catb/trec-web-2009')

        with self.subTest('all fields'):
            topics = dataset.get_topics()
            self.assertEqual(['qid', 'query', 'description', 'type', 'subtopics'], list(topics.columns))

        with self.subTest('specific field'):
            topics = dataset.get_topics('description')
            self.assertEqual(['qid', 'query'], list(topics.columns)) # description mapped to query
            self.assertEqual(topics.iloc[0]['query'], "Find information on President Barack Obama's family\n  history, including genealogy, national origins, places and dates of\n  birth, etc.\n  ")

        with self.subTest('specific field'):
            topics = dataset.get_topics('description')
            self.assertEqual(['qid', 'query'], list(topics.columns)) # description mapped to query
            self.assertEqual(topics.iloc[0]['query'], "Find information on President Barack Obama's family\n  history, including genealogy, national origins, places and dates of\n  birth, etc.\n  ")

        with self.subTest('field named query'):
            topics = dataset.get_topics('query')
            self.assertEqual(['qid', 'query'], list(topics.columns))
            self.assertEqual(topics.iloc[0]['query'], 'obama family tree')

        with self.assertRaises(AssertionError):
            dataset.get_topics('field_that_does_not_exist')

if __name__ == '__main__':
    unittest.main()



import os
import pandas as pd
from typing import Union, Dict, Tuple, Sequence, Literal, Optional, overload, Any
import types

from ._execution import _run_and_evaluate, _precomputation
from ._rendering import EvaluationDataTuple, RenderFromPerQuery
from ._validation import _validate
from . import SYSTEM_OR_RESULTS_TYPE, MEASURES_TYPE, TEST_FN_TYPE, SAVEFORMAT_TYPE, SAVEMODE_TYPE, VALIDATE_TYPE
import pyterrier as pt

# perquery: bool, dataframe:true
@overload
def Experiment(
        retr_systems : Sequence[SYSTEM_OR_RESULTS_TYPE],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        eval_metrics : MEASURES_TYPE,
        names : Optional[Sequence[str]] = None,
        perquery : Union[Literal[False],Literal[True]] = False,
        dataframe : Literal[True] = True,
        batch_size : Optional[int] = None,
        filter_by_qrels : bool = False,
        filter_by_topics : bool = True,
        baseline : Optional[int] = None,
        test : Union[str,TEST_FN_TYPE] = "t",
        correction : Optional[str] = None,
        correction_alpha : float = 0.05,
        highlight : Optional[str] = None,
        round : Optional[Union[int,Dict[str,int]]] = None,
        verbose : bool = False,
        validate : VALIDATE_TYPE = 'warn',
        save_dir : Optional[str] = None,
        save_mode : SAVEMODE_TYPE = 'warn',
        save_format : SAVEFORMAT_TYPE = 'trec',
        precompute_prefix : bool = False,
        **kwargs) -> pd.DataFrame:
    ...

# perquery: bool, dataframe:False
@overload
def Experiment(
        retr_systems : Sequence[SYSTEM_OR_RESULTS_TYPE],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        eval_metrics : MEASURES_TYPE,
        names : Optional[Sequence[str]] = None,
        perquery : Union[Literal[False],Literal[True]] = False,
        dataframe : Literal[False] = False,
        batch_size : Optional[int] = None,
        filter_by_qrels : bool = False,
        filter_by_topics : bool = True,
        baseline : Optional[int] = None,
        test : Union[str,TEST_FN_TYPE] = "t",
        correction : Optional[str] = None,
        correction_alpha : float = 0.05,
        highlight : Optional[str] = None,
        round : Optional[Union[int,Dict[str,int]]] = None,
        verbose : bool = False,
        validate : VALIDATE_TYPE = 'warn',
        save_dir : Optional[str] = None,
        save_mode : SAVEMODE_TYPE = 'warn',
        save_format : SAVEFORMAT_TYPE = 'trec',
        precompute_prefix : bool = False,
        **kwargs) -> Dict[str,Any]:
    ...

# perquery: 'both', dataframe:True
@overload
def Experiment(
        retr_systems : Sequence[SYSTEM_OR_RESULTS_TYPE],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        eval_metrics : MEASURES_TYPE,
        names : Optional[Sequence[str]] = None,
        perquery : Literal['both'] = 'both',
        dataframe : Literal[True] = True,
        batch_size : Optional[int] = None,
        filter_by_qrels : bool = False,
        filter_by_topics : bool = True,
        baseline : Optional[int] = None,
        test : Union[str,TEST_FN_TYPE] = "t",
        correction : Optional[str] = None,
        correction_alpha : float = 0.05,
        highlight : Optional[str] = None,
        round : Optional[Union[int,Dict[str,int]]] = None,
        verbose : bool = False,
        validate : VALIDATE_TYPE = 'warn',
        save_dir : Optional[str] = None,
        save_mode : SAVEMODE_TYPE = 'warn',
        save_format : SAVEFORMAT_TYPE = 'trec',
        precompute_prefix : bool = False,
        **kwargs) -> Tuple[pd.DataFrame,pd.DataFrame]:
    ...

# perquery: 'both', dataframe:False
@overload
def Experiment(
        retr_systems : Sequence[SYSTEM_OR_RESULTS_TYPE],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        eval_metrics : MEASURES_TYPE,
        names : Optional[Sequence[str]] = None,
        perquery : Literal['both'] = 'both',
        dataframe : Literal[False] = False,
        batch_size : Optional[int] = None,
        filter_by_qrels : bool = False,
        filter_by_topics : bool = True,
        baseline : Optional[int] = None,
        test : Union[str,TEST_FN_TYPE] = "t",
        correction : Optional[str] = None,
        correction_alpha : float = 0.05,
        highlight : Optional[str] = None,
        round : Optional[Union[int,Dict[str,int]]] = None,
        verbose : bool = False,
        validate : VALIDATE_TYPE = 'warn',
        save_dir : Optional[str] = None,
        save_mode : SAVEMODE_TYPE = 'warn',
        save_format : SAVEFORMAT_TYPE = 'trec',
        precompute_prefix : bool = False,
        **kwargs) -> Tuple[Dict[str,Any], Dict[str,Any]]:
    ...

def Experiment(
        retr_systems : Sequence[SYSTEM_OR_RESULTS_TYPE],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        eval_metrics : MEASURES_TYPE,
        names : Optional[Sequence[str]] = None,
        perquery : Union[bool, Literal['both']] = False,
        dataframe : Union[Literal[False], Literal[True]] = True,
        batch_size : Optional[int] = None,
        filter_by_qrels : bool = False,
        filter_by_topics : bool = True,
        baseline : Optional[int] = None,
        test : Union[str,TEST_FN_TYPE] = "t",
        correction : Optional[str] = None,
        correction_alpha : float = 0.05,
        highlight : Optional[str] = None,
        round : Optional[Union[int,Dict[str,int]]] = None,
        verbose : bool = False,
        validate : VALIDATE_TYPE = 'warn',
        save_dir : Optional[str] = None,
        save_mode : SAVEMODE_TYPE = 'warn',
        save_format : SAVEFORMAT_TYPE = 'trec',
        precompute_prefix : bool = False,
        **kwargs):
    """
    Allows easy comparison of multiple retrieval transformer pipelines using a common set of topics, and
    identical evaluation measures computed using the same qrels. In essence, each transformer is applied on 
    the provided set of topics. Then the named evaluation measures are computed for each system.

    :param retr_systems: A list of transformers to evaluate. If you already have the results for one 
        (or more) of your systems, a results dataframe can also be used here. Results produced by 
        the transformers must have "qid", "docno", "score", "rank" columns.
    :param topics: Either a path to a topics file or a pandas.Dataframe with columns=['qid', 'query']
    :param qrels: Either a path to a qrels file or a pandas.Dataframe with columns=['qid','docno', 'label']   
    :param eval_metrics: Which evaluation metrics to use. E.g. ['map']
    :param names: List of names for each retrieval system when presenting the results.
        Default=None. If None: Obtains the `str()` representation of each transformer as its name.
    :param batch_size: If not None, evaluation is conducted in batches of batch_size topics. Default=None, which evaluates all topics at once. 
        Applying a batch_size is useful if you have large numbers of topics, and/or if your pipeline requires large amounts of temporary memory
        during a run.
    :param filter_by_qrels: If True, will drop topics from the topics dataframe that have qids not appearing in the qrels dataframe. 
    :param filter_by_topics: If True, will drop topics from the qrels dataframe that have qids not appearing in the topics dataframe. 
    :param perquery: If True return each metric for each query, if False, will return mean metrics across all queries. If both, will return both averages and perquery results in a tuple. Default=False.
    :param save_dir: If set to the name of a directory, the results of each transformer will be saved in TREC-formatted results file, whose 
        filename is based on the systems names (as specified by ``names`` kwarg). If the file exists and ``save_mode`` is set to "reuse", then the file
        will be used for evaluation rather than the transformer. Default is None, such that saving and loading from files is disabled.
    :param save_mode: Defines how existing files are used when ``save_dir`` is set. If set to "reuse", then files will be preferred
        over transformers for evaluation. If set to "overwrite", existing files will be replaced. If set to "warn" or "error", the presence of any 
        existing file will cause a warning or error, respectively. Default is "warn".
    :param save_format: How are result being saved. Defaults to 'trec', which uses ``pt.io.read_results()`` and ``pt.io.write_results()`` for saving system outputs. 
        If TREC results format is insufficient, set ``save_format=pickle``. Alternatively, a tuple of read and write function can be specified, for instance, 
        ``save_format=(pandas.from_csv, pandas.DataFrame.to_csv)``, or even ``save_format=(pandas.from_parquet, pandas.DataFrame.to_parquet)``.
    :param dataframe: If True return results as a dataframe, else as a dictionary of dictionaries. Default=True.
    :param baseline: If set to the index of an item of the retr_system list, will calculate the number of queries 
        improved, degraded and the statistical significance (paired t-test p value) for each measure.
        Default=None: If None, no additional columns will be added for each measure.
    :param test: Which significance testing approach to apply. Defaults to "t". Alternatives are "wilcoxon" - not typically used for IR experiments. A Callable can also be passed - it should
        follow the specification of `scipy.stats.ttest_rel() <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ttest_rel.html>`_, 
        i.e. it expect two arrays of numbers, and return an array or tuple, of which the second value will be placed in the p-value column.
    :param correction: Whether any multiple testing correction should be applied. E.g. 'bonferroni', 'holm', 'hs' aka 'holm-sidak'. Default is None.
        Additional columns are added denoting whether the null hypothesis can be rejected, and the corrected p value. 
        See `statsmodels.stats.multitest.multipletests() <https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html#statsmodels.stats.multitest.multipletests>`_
        for more information about available testing correction.
    :param correction_alpha: What alpha value for multiple testing correction. Default is 0.05.
    :param highlight: If `highlight="bold"`, highlights in bold the best measure value in each column; 
        if `highlight="color"` or `"colour"`, then the cell with the highest metric value will have a green background.
    :param round: How many decimal places to round each measure value to. This can also be a dictionary mapping measure name to number of decimal places.
        Default is None, which is no rounding.
    :param precompute_prefix: If set to True, then pt.Experiment will look for a common prefix on all input pipelines, and execute that common prefix pipeline only once. 
        This functionality assumes that the intermidiate results of the common prefix can fit in memory. Set to False by default.
    :param verbose: If True, a tqdm progress bar is shown as systems (or systems*batches if batch_size is set) are executed. Default=False.
    :param validate: If set to value other than 'ignore', each transformer is validated against the topics dataframe, to ensure that it produces the expected output columns.
        ``pt.inspect.transformer_outputs()`` is used to determine the output columns. If 'warn', then transformers whose output columns don't match the columns required 
        by the specified evaluation measures will product warnings; If 'error', then an error is produced. If a transformer cannot be inspected, a warning is produced.

    :return: A Dataframe/dict with each retrieval system with each metric evaluated, or alternatively a tuple with averages and perquery results. 
    """
    
    if not isinstance(retr_systems, list):
        raise TypeError("Expected list of transformers for retr_systems, instead received %s" % str(type(retr_systems)))

    if len(kwargs):
        raise TypeError("Unknown kwargs: %s" % (str(list(kwargs.keys()))))

    if baseline is not None:
        assert int(baseline) >= 0 and int(baseline) < len(retr_systems)
        assert not perquery

    if isinstance(topics, str):
        if os.path.isfile(topics):
            topics = pt.io.read_topics(topics)
    if isinstance(qrels, str):
        if os.path.isfile(qrels):
            qrels = pt.io.read_qrels(qrels)

    if round is not None:
        if isinstance(round, int):
            assert round >= 0, "round argument should be integer >= 0, not %s" % str(round)
        elif isinstance(round, dict):
            assert not perquery, "Sorry, per-measure rounding only support when reporting means" 
            for k,v in round.items():
                assert isinstance(v, int) and v >= 0, "rounding number for measure %s should be integer >= 0, not %s" % (k, str(v))
        else:
            raise ValueError("Argument round should be an integer or a dictionary")

    if correction is not None and baseline is None:
        raise ValueError("Requested multiple testing correction, but no baseline was specified.")

    # drop queries not appear in the qrels
    if filter_by_qrels:
        # the commented variant would drop queries not having any RELEVANT labels
        # topics = topics.merge(qrels[qrels["label"] > 0][["qid"]].drop_duplicates())        
        topics = topics.merge(qrels[["qid"]].drop_duplicates())
        if len(topics) == 0:
            raise ValueError('There is no overlap between the qids found in the topics and qrels. If this is intentional, set filter_by_topics=False and filter_by_qrels=False.')

    # drop qrels not appear in the topics
    if filter_by_topics:
        qrels = qrels.merge(topics[["qid"]].drop_duplicates())
        if len(qrels) == 0:
            raise ValueError('There is no overlap between the qids found in the topics and qrels. If this is intentional, set filter_by_topics=False and filter_by_qrels=False.')

    from scipy import stats
    test_fn : TEST_FN_TYPE
    if test == "t":
        test_fn = stats.ttest_rel
    elif test == "wilcoxon":
        test_fn = stats.wilcoxon
    else:
        assert not isinstance(test, str), "Unknown test function name %s" % test
        test_fn = test
    
    # obtain system names if not specified
    if names is None:
        names = [str(system) for system in retr_systems]
    elif len(names) != len(retr_systems):
        raise ValueError("names should be the same length as retr_systems")

    # validate save_dir and resulting filenames
    if save_dir is not None:
        if not os.path.exists(save_dir):
            raise ValueError("save_dir %s does not exist" % save_dir)
        if not os.path.isdir(save_dir):
            raise ValueError("save_dir %s is not a directory" % save_dir)
        from ..io import ok_filename
        for n in names:
            if not ok_filename(n):
                raise ValueError("Name contains bad characters and save_dir is set, name is %s" % n)
        if len(set(names)) < len(names):
            raise ValueError("save_dir is set, but names are not unique. Use names= to set unique names")

    all_topic_qids = topics["qid"].values

    mrt_needed = False
    if "mrt" in eval_metrics:
        mrt_needed = True
        eval_metrics = list(eval_metrics).copy()
        eval_metrics.remove("mrt")

    # validate the transformers produce the expected columns
    _validate(retr_systems, topics, eval_metrics, names, validate)

    # split the transformers into a common prefix and individual suffixes, improved efficiency
    precompute_time, execution_topics, execution_retr_systems = _precomputation(retr_systems, topics, precompute_prefix, verbose, batch_size)

    # progress bar construction
    tqdm_args={
        'disable' : not verbose,
        'unit' : 'system',
        'total' : len(retr_systems),
        'desc' : 'pt.Experiment'
    }

    if batch_size is not None:
        import math
        tqdm_args['unit'] = 'batches'
        # round number of batches up for each system
        tqdm_args['total'] = math.ceil((len(topics) / batch_size)) * len(retr_systems)

    renderer = RenderFromPerQuery(names, 
                                  baseline=baseline, 
                                  test_fn=test_fn, 
                                  correction=correction, 
                                  correction_alpha=correction_alpha, 
                                  round=round, 
                                  precompute_time=precompute_time)
    with pt.tqdm(**tqdm_args) as pbar:
        # run and evaluate each system
        for sysid, (name, system) in enumerate(zip(names, execution_retr_systems)):
            save_file = None
            if save_dir is not None:
                if save_format == 'trec':
                    save_ext = 'res.gz'
                elif isinstance(save_format, types.ModuleType):
                    save_ext = 'mod'
                elif isinstance(save_format, tuple):
                    save_ext = 'custom'
                else:
                    raise ValueError("Unrecognised save_mode %s" % str(save_format)) 
                save_file = os.path.join(save_dir, "%s.%s" % (name, save_ext))

            time, evalMeasuresDict = _run_and_evaluate(
                system, execution_topics, qrels, eval_metrics, 
                perquery=True, 
                batch_size=batch_size, 
                backfill_qids=all_topic_qids if perquery else None,
                save_file=save_file,
                save_mode=save_mode,
                save_format=save_format,
                pbar=pbar)
            renderer.add_metrics(sysid, evalMeasuresDict, time)

    if not perquery:
        return renderer.averages(dataframe=dataframe, highlight=highlight, mrt_needed=mrt_needed)
    
    perquery_results = renderer.perquery(dataframe=dataframe)
    if perquery == 'both':
        average_results = renderer.averages(dataframe=dataframe, highlight=highlight)

        if dataframe:
            from typing import cast as tcast
            average_results_df = tcast(pd.DataFrame, average_results)
            perquery_results_df = tcast(pd.DataFrame, perquery_results)

            average_results_df.style.set_caption("Averages")
            perquery_results_df.style.set_caption("Per Query")
            return EvaluationDataTuple(average_results, perquery_results)
        return (average_results, perquery_results)
    return perquery_results



from enum import Enum
from typing import Union, List, Dict
import pyterrier as pt


class TerrierStopwords(Enum):
    """The stopword configuration to use for Terrier.

    This enum is primarily used with :class:`~pyterrier.terrier.TerrierIndexer.indexer`.
    """

    none = 'none' #: No Stopwords
    terrier = 'terrier' #: Apply Terrier's standard stopword list
    custom = 'custom' #: Apply PyTerrierCustomStopwordList.Indexing for indexing, and PyTerrierCustomStopwordList.Retrieval for retrieval

    @staticmethod
    def _to_obj(this):
        if isinstance(this, list):
            rtr = TerrierStopwords('custom')
            return rtr, list(this)
        try:
            return TerrierStopwords(this), None
        except ValueError:
            return this, None
        
    @staticmethod
    @pt.java.required
    def _indexing_config(this, stopword_list : Union[List[str], None], termpipelines : List[str], properties : Dict[str,str], hooks : List):
        if this is None or this == TerrierStopwords.none:
            pass
        if this == TerrierStopwords.terrier:
            termpipelines.append('Stopwords')
        if this == TerrierStopwords.custom:
            cst = pt.java.cast
            assert pt.terrier.check_version("5.8"), "Terrier 5.8 required"
            assert stopword_list is not None, "expected to receive a stopword list"

            stopword_list_esc = [t.replace(",", "\\,") for t in stopword_list ]

            properties["pyterrier.stopwords"]  = ",".join(stopword_list_esc)
            termpipelines.append('org.terrier.python.PyTerrierCustomStopwordList$Indexing')

            # this hook updates the index's properties to handle the python stopwords list
            def _hook(pyindexer, index):
                pindex = cst("org.terrier.structures.PropertiesIndex", index)
                # store the stopwords into the Index's properties
                pindex.setIndexProperty("pyterrier.stopwords", ",".join(stopword_list_esc))

                # change the stopwords list implementation: the Indexing variant obtains
                # stopwords from the global ApplicationSetup properties, while the 
                # Retrieval variant obtains them from the *Index* properties instead
                pindex.setIndexProperty("termpipelines", 
                    pindex.getIndexProperty('termpipelines', None)
                    .replace('org.terrier.python.PyTerrierCustomStopwordList$Indexing',
                             'org.terrier.python.PyTerrierCustomStopwordList$Retrieval'))
                pindex.flush()
            hooks.append(_hook)



from typing import Union, Optional, Callable, List, Any
import pandas as pd
import numpy as np
import re
from warnings import warn
from pyterrier.datasets import Dataset
from pyterrier.model import FIRST_RANK
import concurrent
from concurrent.futures import ThreadPoolExecutor
import pyterrier as pt
from typing import Dict

from pyterrier.terrier.stemmer import TerrierStemmer
from pyterrier.terrier.stopwords import TerrierStopwords
from pyterrier.terrier.tokeniser import TerrierTokeniser

def _query_needs_tokenised(query: str) -> bool:
    # detects if we need to tokenise the query
    if _matchop(query):
        return False
    
    if "applypipeline" in query:
        return False
    
    # does it contains in terrier language 
    termweightsre = re.compile(r'\^\d+(\.\d+)?')
    if termweightsre.match(query):
        return False

    # we dont include : in this list, as it denotes a control key:value pair in TerrierQL
    bad_chars = [",", ".", ";", "!", "?", "(", ")", "[", "]", "{", "}", "<", ">", "\"", "^", 
                 "'", "`", "~", "@", "#", "$", "%", "&", "*", "-", "+", "=", "|", "\\", "/"]

    if any(c in query for c in bad_chars):
        return True
    
    return False

_matchops = ["#combine", "#uw", "#1", "#tag", "#prefix", "#band", "#base64", "#syn"]

def _matchop(query):
    for m in _matchops:
        if m in query:
            return True
    return False

def _querytoks2matchop(query_toks: Dict[str,float]) -> str:
    def _matchop_tok(t, w):
        import base64
        import string
        if not all(a in string.ascii_letters + string.digits for a in t):
            encoded = base64.b64encode(t.encode('utf-8')).decode("utf-8") 
            t = f'#base64({encoded})'
        if w != 1:
            t = f'#combine:0={w:f}({t})'
        return t
    return ' '.join([ _matchop_tok(t, w) for (t,w) in query_toks.items() ])

@pt.java.required
def _function2wmodel(function):
    from jnius import PythonJavaClass, java_method

    @pt.java.required
    class PythonWmodelFunction(PythonJavaClass):
        __javainterfaces__ = ['org/terrier/python/CallableWeightingModel$Callback']

        def __init__(self, fn):
            super(PythonWmodelFunction, self).__init__()
            self.fn = fn
            
        @java_method('(DLorg/terrier/structures/postings/Posting;Lorg/terrier/structures/EntryStatistics;Lorg/terrier/structures/CollectionStatistics;)D', name='score')
        def score(self, keyFreq, posting, entryStats, collStats):
            return self.fn(keyFreq, posting, entryStats, collStats)

        @java_method('()Ljava/nio/ByteBuffer;')
        def serializeFn(self):
            import dill as pickle
            #see https://github.com/SeldonIO/alibi/issues/447#issuecomment-881552005
            from dill import extend
            extend(use_dill=False)
            # keep both python and java representations around to prevent them being GCd in respective VMs 
            self.pbyterep = pickle.dumps(self.fn)
            self.jbyterep = pt.java.autoclass("java.nio.ByteBuffer").wrap(self.pbyterep)
            return self.jbyterep

    callback = PythonWmodelFunction(function)
    wmodel = pt.java.autoclass("org.terrier.python.CallableWeightingModel")( callback )
    return callback, wmodel

def _mergeDicts(defaults : Dict[str,str], settings : Optional[Dict[str,str]] = None) -> Dict[str,str]:
    KV = defaults.copy()
    if settings is not None and len(settings) > 0:
        KV.update(settings)
    return KV

@pt.java.required
def _parse_index_like(index_location):
    JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
    JI = pt.java.autoclass('org.terrier.structures.Index')
    from pyterrier.terrier import TerrierIndexer

    if isinstance(index_location, JIR):
        return index_location
    if isinstance(index_location, JI):
        return pt.java.cast('org.terrier.structures.Index', index_location).getIndexRef()
    if isinstance(index_location, str) or issubclass(type(index_location), TerrierIndexer):
        if issubclass(type(index_location), TerrierIndexer):
            return JIR.of(index_location.path)
        return JIR.of(index_location)

    raise ValueError(
        f'''index_location is current a {type(index_location)},
        while it needs to be an Index, an IndexRef, a string that can be
        resolved to an index location (e.g. path/to/index/data.properties),
        or an pyterrier.index.TerrierIndexer object'''
    )

  
@pt.java.required
class Retriever(pt.Transformer):
    """
    Use this class for retrieval by Terrier
    """

    @staticmethod
    def matchop(t, w=1):
        """
        Static method used for rewriting a query term to use a MatchOp operator if it contains
        anything except ASCII letters or digits.
        """
        import base64
        import string
        if not all(a in string.ascii_letters + string.digits for a in t):
            encoded = base64.b64encode(t.encode('utf-8')).decode("utf-8") 
            t = f'#base64({encoded})'
        if w != 1:
            t = f'#combine:0={w}({t})'
        return t


    @staticmethod
    def from_dataset(dataset : Union[str,Dataset], 
            variant : Optional[str] = None, 
            version='latest',            
            **kwargs):
        """
        Static method that instantiates a Retriever object from a pre-built index access via a dataset.
        Pre-built indices are ofen provided via the `Terrier Data Repository <http://data.terrier.org/>`_.

        :param dataset: The name of the dataset, or a Dataset object that contains the index
        :param variant: The variant of the index to use. If None, the default index will be used.
        :param version: The version of the dataset to use. If None, the latest version will be used.

        Examples::

            dataset = pt.get_dataset("vaswani")
            bm25 = pt.terrier.Retriever.from_dataset(dataset, "terrier_stemmed", wmodel="BM25")
            #or
            bm25 = pt.terrier.Retriever.from_dataset("vaswani", "terrier_stemmed", wmodel="BM25")

        **Index Variants**:

        There are a number of standard index names.
         - `terrier_stemmed` - a classical index, removing Terrier's standard stopwords, and applying Porter's English stemmer
         - `terrier_stemmed_positions` - as per `terrier_stemmed`, but also containing position information
         - `terrier_unstemmed` - a classical index, without applying stopword removal or stemming
         - `terrier_stemmed_text` - as per `terrier_stemmed`, but also containing the raw text of the documents
         - `terrier_unstemmed_text` - as per `terrier_stemmed`, but also containing the raw text of the documents

        """
        return pt.datasets.transformer_from_dataset(dataset, variant=variant, version=version, clz=Retriever, **kwargs)

    #: default_controls(dict): stores the default controls
    default_controls = {
        "terrierql": "on",
        "parsecontrols": "on",
        "parseql": "on",
        "applypipeline": "on",
        "localmatching": "on",
        "filters": "on",
        "decorate": "on",
        "wmodel": "DPH",
    }

    #: default_properties(dict): stores the default properties
    default_properties = {
        "querying.processes": "terrierql:TerrierQLParser,parsecontrols:TerrierQLToControls,parseql:TerrierQLToMatchingQueryTerms,matchopql:MatchingOpQLParser,applypipeline:ApplyTermPipeline,context_wmodel:org.terrier.python.WmodelFromContextProcess,localmatching:LocalManager$ApplyLocalMatching,qe:QueryExpansion,labels:org.terrier.learning.LabelDecorator,filters:LocalManager$PostFilterProcess",
        "querying.postfilters": "decorate:SimpleDecorate,site:SiteFilter,scope:Scope",
        "querying.default.controls": "wmodel:DPH,parsecontrols:on,parseql:on,applypipeline:on,terrierql:on,localmatching:on,filters:on,decorate:on",
        "querying.allowed.controls": "scope,qe,qemodel,start,end,site,scope,applypipeline",
        "termpipelines": "Stopwords,PorterStemmer"
    }

    def __init__(self, 
                 index_location : Union[str,Any], 
                 controls : Optional[Dict[str,str]] = None, 
                 properties : Optional[Dict[str,str]]= None, 
                 metadata : List[str] = ["docno"], 
                 num_results : Optional[int] = None, 
                 wmodel : Optional[Union[str, Callable]] = None, 
                 tokeniser : Union[str,TerrierTokeniser] = TerrierTokeniser.english,
                 threads : int = 1, 
                 verbose : bool = False):
        """
            Init method

            :param index_location: An index-like object - An Index, an IndexRef, or a String that can be resolved to an IndexRef
            :param controls: A dictionary with the control names and values
            :param properties: A dictionary with the property keys and values
            :param verbose: If True transform method will display progress
            :param num_results: Number of results to retrieve. 
            :param metadata: What metadata to retrieve. Default is ["docno"]. 
        """
        self.indexref = _parse_index_like(index_location)
        self.properties = _mergeDicts(Retriever.default_properties, properties)
        self.concurrentIL = pt.java.autoclass("org.terrier.structures.ConcurrentIndexLoader")
        if pt.terrier.check_version(5.5) and "SimpleDecorateProcess" not in self.properties["querying.processes"]:
            self.properties["querying.processes"] += ",decorate:SimpleDecorateProcess"
            controls = controls or {}
            controls["decorate_batch"] = "on"
        self.metadata = metadata
        self.threads = threads
        self.RequestContextMatching = pt.java.autoclass("org.terrier.python.RequestContextMatching")
        self.search_context = {}
        self.verbose = verbose
        self.tokeniser = TerrierTokeniser.java_tokeniser(TerrierTokeniser._to_obj(tokeniser))

        for key, value in self.properties.items():
            pt.terrier.J.ApplicationSetup.setProperty(str(key), str(value))
        
        self.controls = _mergeDicts(Retriever.default_controls, controls)
        if wmodel is not None:
            from pyterrier.transformer import is_lambda, is_function
            if isinstance(wmodel, str):
                self.controls["wmodel"] = wmodel
            elif is_lambda(wmodel) or is_function(wmodel):
                callback, wmodelinstance = _function2wmodel(wmodel)
                #save the callback instance in this object to prevent being GCd by Python
                self._callback = callback
                self.search_context['context_wmodel'] = wmodelinstance
                self.controls['context_wmodel'] = 'on'
            elif isinstance(wmodel, pt.java.autoclass("org.terrier.matching.models.WeightingModel")):
                self.search_context['context_wmodel'] = wmodel
                self.controls['context_wmodel'] = 'on'
            else:
                raise ValueError("Unknown parameter type passed for wmodel argument: %s" % str(wmodel))
                  
        if self.threads > 1:
            warn(
                "Multi-threaded retrieval is experimental, YMMV.")
            assert pt.terrier.check_version(5.5), "Terrier 5.5 is required for multi-threaded retrieval"

            # we need to see if our indexref is concurrent. if not, we upgrade it using ConcurrentIndexLoader
            # this will upgrade the underlying index too.
            if not self.concurrentIL.isConcurrent(self.indexref):
                warn(
                    "Upgrading indexref %s to be concurrent" % self.indexref.toString())
                self.indexref = self.concurrentIL.makeConcurrent(self.indexref)

        if num_results is not None:
            if num_results > 0:
                self.controls["end"] = str(num_results -1)
            elif num_results == 0:
                del self.controls["end"]
            else: 
                raise ValueError("num_results must be None, 0 or positive")


        MF = pt.java.autoclass('org.terrier.querying.ManagerFactory')
        self.RequestContextMatching = pt.java.autoclass("org.terrier.python.RequestContextMatching")
        self.manager = MF._from_(self.indexref)
    
    def get_parameter(self, name : str):
        if name in self.controls:
            return self.controls[name]
        elif name in self.properties:
            return self.properties[name]
        else:
            return super().get_parameter(name)

    def set_parameter(self, name : str, value):
        if name in self.controls:
            self.controls[name] = value
        elif name in self.properties:
            self.properties[name] = value
        else:
            super().set_parameter(name,value)

    def __reduce__(self):
        return (
            self.__class__,
            (self.indexref,),
            self.__getstate__()
        )

    def __getstate__(self): 
        return  {
                'context' : self.search_context,
                'controls' : self.controls, 
                'properties' : self.properties, 
                'metadata' : self.metadata,
                }

    def __setstate__(self, d): 
        self.controls = d["controls"]
        self.metadata = d["metadata"]
        self.search_context = d["context"]
        self.properties.update(d["properties"])
        for key,value in d["properties"].items():
            pt.terrier.J.ApplicationSetup.setProperty(key, str(value))

    def _retrieve_one(self, row, input_results=None, docno_provided=False, docid_provided=False, scores_provided=False):
        rank = FIRST_RANK
        qid = str(row.qid)

        # row is a namedtuple, whose fields are exposed in _fields
        query_toks_present = 'query_toks' in row._fields
        if query_toks_present:
            query = '' # Clear the query so it doesn't match the "applypipeline:off" or "_matchop" condictions below... The query_toks query is converted below.
            srq = self.manager.newSearchRequest(qid)
        else:
            query = row.query
            if _query_needs_tokenised(query):
                query = ' '.join(self.tokeniser.getTokens(query))
            if len(query) == 0:
                warn(
                    "Skipping empty query for qid %s" % qid)
                return []
            srq = self.manager.newSearchRequest(qid, query)

        for control, value in self.controls.items():
            srq.setControl(control, str(value))

        for key, value in self.search_context.items():
            srq.setContextObject(key, value)

        # this is needed until terrier-core issue #106 lands
        if "applypipeline:off" in query:
            srq.setControl("applypipeline", "off")
            srq.setOriginalQuery(query.replace("applypipeline:off", ""))

        # transparently detect matchop queries
        if _matchop(query):
            srq.setControl("terrierql", "off")
            srq.setControl("parsecontrols", "off")
            srq.setControl("parseql", "off")
            srq.setControl("matchopql", "on")

        if query_toks_present:
            if len(row.query_toks) == 0:
                warn(
                    "Skipping empty query_toks for qid %s" % qid)
                return []
            srq.setControl("terrierql", "off")
            srq.setControl("parsecontrols", "off")
            srq.setControl("parseql", "off")
            srq.setControl("matchopql", "on")
            query = _querytoks2matchop(row.query_toks)
            srq.setOriginalQuery(query)

        #ask decorate only to grab what we need
        srq.setControl("decorate", ",".join(self.metadata))

        # this handles the case that a candidate set of documents has been set. 
        num_expected = None
        if docno_provided or docid_provided:
            # we use RequestContextMatching to make a ResultSet from the 
            # documents in the candidate set. 
            matching_config_factory = self.RequestContextMatching.of(srq)
            input_query_results = input_results[input_results["qid"] == qid]
            num_expected = len(input_query_results)
            if docid_provided:
                matching_config_factory.fromDocids(input_query_results["docid"].values.tolist())
            elif docno_provided:
                matching_config_factory.fromDocnos(input_query_results["docno"].values.tolist())
            # batch retrieve is a scoring process that always overwrites the score; no need to provide scores as input
            #if scores_provided:
            #    matching_config_factory.withScores(input_query_results["score"].values.tolist())
            matching_config_factory.build()
            srq.setControl("matching", "org.terrier.matching.ScoringMatching" + "," + srq.getControl("matching"))

        # now ask Terrier to run the request
        self.manager.runSearchRequest(srq)
        result = srq.getResults()

        # check we got all of the expected metadata (if the resultset has a size at all)
        if len(result) > 0 and len(set(self.metadata) & set(result.getMetaKeys())) != len(self.metadata):
            raise KeyError("Mismatch between requested and available metadata in %s. Requested metadata: %s, available metadata %s" % 
                (str(self.indexref), str(self.metadata), str(result.getMetaKeys()))) 

        if num_expected is not None:
            assert(num_expected == len(result))
        
        rtr_rows=[]
        # prepare the dataframe for the results of the query
        for item in result:
            metadata_list = []
            for meta_column in self.metadata:
                metadata_list.append(item.getMetadata(meta_column))
            res = [qid, item.getDocid()] + metadata_list + [rank, item.getScore()]
            rank += 1
            rtr_rows.append(res)

        return rtr_rows


    def transform(self, queries):
        """
        Performs the retrieval

        :param queries: a pandas.Dataframe with columns=['qid', 'query']. For re-ranking,
                the DataFrame may also have a 'docid' and or 'docno' column.

        :return: pandas.Dataframe with columns=['qid', 'docno', 'rank', 'score']
        """
        results=[]
        if not isinstance(queries, pd.DataFrame):
            raise ValueError(".transform() should be passed a DataFrame (found %s). Use .search() to execute a single query; Use .transform_iter() for iter-dicts" % str(type(queries)))
        
        # use pt.validate - this makes inspection of input columns better
        with pt.validate.any(queries) as v:
            v.columns(includes=['qid', 'query'], excludes=['docid', 'docno'], mode='retrieve') # query based frame without docno or docid
            v.columns(includes=['qid', 'query', 'docid'], mode='rerank') # docid-based results frame
            v.query_frame(extra_columns=['query_toks'], mode='retrieve_toks')
            v.result_frame(extra_columns=['query'], mode='rerank')
            
        docno_provided = "docno" in queries.columns
        docid_provided = "docid" in queries.columns
        scores_provided = "score" in queries.columns
        input_results = None
        if v.mode == 'rerank':
            assert docno_provided or docid_provided, "For reranking, either docno or docid must be provided"
            assert pt.terrier.check_version(5.3)
            input_results = queries

            # query is optional, and functionally dependent on qid.
            # Hence as long as one row has the query for each qid, 
            # the rest can be None
            queries = input_results[["qid", "query"]].dropna(axis=0, subset=["query"]).drop_duplicates()
            
        # make sure queries are a String
        if queries["qid"].dtype == np.int64:
            queries['qid'] = queries['qid'].astype(str)

        if self.threads > 1:

            if not self.concurrentIL.isConcurrent(self.indexref):
                raise ValueError("Threads must be set >1 in constructor and/or concurrent indexref used")
            
            with ThreadPoolExecutor(max_workers=self.threads) as executor:
                
                # we must detatch jnius to prevent thread leaks through JNI
                from jnius import detach
                def _one_row(*args, **kwargs):
                    rtr = self._retrieve_one(*args, **kwargs)
                    detach()
                    return rtr
                
                # create a future for each query, and submit to Terrier
                future_results = {
                    executor.submit(_one_row, row, input_results, docno_provided=docno_provided, docid_provided=docid_provided, scores_provided=scores_provided) : row.qid 
                    for row in queries.itertuples()}
                
                # as these futures complete, wait and add their results
                iter = concurrent.futures.as_completed(future_results)
                if self.verbose and len(queries):
                    iter = pt.tqdm(iter, desc=str(self), total=queries.shape[0], unit="q")
                
                for future in iter:
                    res = future.result()
                    results.extend(res)
        else:
            iter = queries.itertuples()
            if self.verbose and len(queries):
                iter = pt.tqdm(iter, desc=str(self), total=queries.shape[0], unit="q")
            for row in iter:
                res = self._retrieve_one(row, input_results, docno_provided=docno_provided, docid_provided=docid_provided, scores_provided=scores_provided)
                results.extend(res)

        res_dt = pd.DataFrame(results, columns=['qid', 'docid' ] + self.metadata + ['rank', 'score'])
        # ensure to return the query and any other input columns
        input_cols = queries.columns[ (queries.columns == "qid") | (~queries.columns.isin(res_dt.columns))]
        res_dt = res_dt.merge(queries[input_cols], on=["qid"])
        return res_dt
        
    def __repr__(self):
        return "TerrierRetr(" + ",".join([
            self.indexref.toString(),
            str(self.controls),
            str(self.properties)
            ]) + ")"

    def __str__(self):
        return "TerrierRetr(" + self.controls["wmodel"] + ")"

    def setControls(self, controls):
        for key, value in controls.items():
            self.controls[str(key)] = str(value)

    def setControl(self, control, value):
        self.controls[str(control)] = str(value)

    def schematic(self, *, input_columns = None): 
        return {'label': self.controls['wmodel']}

    def fuse_rank_cutoff(self, k: int) -> Optional[pt.Transformer]:
        """
        Support fusing with RankCutoffTransformer.
        """
        if float(self.controls.get('end', float('inf'))) < k:
            return self # the applied rank cutoff is greater than the one already applied
        if self.controls.get('context_wmodel') == 'on':
            return None # we don't store the original wmodel value so we can't reconstruct
        # apply the new k as num_results
        return pt.inspect.transformer_apply_attributes(self, num_results=k)

    def fuse_feature_union(self, other: pt.Transformer, is_left: bool) -> Optional[pt.Transformer]:
        if isinstance(other, Retriever) and \
           self.indexref == other.indexref and \
           self.controls.get('context_wmodel') != 'on' and \
           other.controls.get('context_wmodel') != 'on':
            features = ["WMODEL:" + self.controls['wmodel'], "WMODEL:" + other.controls['wmodel']] if is_left else ["WMODEL:" + other.controls['wmodel'], "WMODEL:" + self.controls['wmodel']]
            controls = dict(self.controls)
            del controls['wmodel']
            return FeaturesRetriever(self.indexref, features, controls=controls, properties=self.properties,
                metadata=self.metadata, threads=self.threads, verbose=self.verbose)
        return None

    def attributes(self):
        if 'wmodel' in self.controls:
            wmodel = self.controls['wmodel']
        elif self.controls.get('context_wmodel') == 'on':
            wmodel = self.search_context['context_wmodel']
        return [
            pt.inspect.TransformerAttribute('index_location', self.indexref),
            pt.inspect.TransformerAttribute('num_results', int(self.controls.get('end', '999')) + 1),
            pt.inspect.TransformerAttribute('metadata', self.metadata),
            pt.inspect.TransformerAttribute('wmodel', wmodel),
            pt.inspect.TransformerAttribute('threads', self.threads),
            pt.inspect.TransformerAttribute('verbose', self.verbose),
        ] + [
            pt.inspect.TransformerAttribute(key, value)
            for key, value in self.controls.items()
            if key not in ('end', 'wmodel')
        ] + [
            pt.inspect.TransformerAttribute(key, value)
            for key, value in self.properties.items()
        ]

    def apply_attributes(self, **kwargs):
        for attr in self.attributes():
            if attr.name not in kwargs:
                kwargs[attr.name] = attr.value
        kwargs['controls'] = {}
        kwargs['properties'] = {}
        for key, value in list(kwargs.items()):
            if key in ('index_location', 'controls', 'properties', 'metadata', 'num_results', 'wmodel', 'threads', 'verbose'):
                pass
            elif key in self.properties:
                kwargs['properties'][key] = value
                del kwargs[key]
            else:
                kwargs['controls'][key] = value
                del kwargs[key]
        return Retriever(**kwargs)


@pt.java.required
class TextIndexProcessor(pt.Transformer):
    '''
        Creates a new MemoryIndex based on the contents of documents passed to it.
        It then creates a new instance of the innerclass and passes the topics to that.

        This class is the base class for TextScorer, but can be used in other settings as well, 
        for instance query expansion based on text.
    '''

    def __init__(self, innerclass, 
                 takes="queries", 
                 returns="docs", 
                 body_attr="body", 
                 background_index=None, 
                 stemmer : Union[None, str, TerrierStemmer] = TerrierStemmer.porter,
                 stopwords : Union[None, TerrierStopwords, List[str]] = TerrierStopwords.terrier,
                 tokeniser : Union[str,TerrierTokeniser] = TerrierTokeniser.english,
                 verbose=False, 
                 **kwargs):
        #super().__init__(**kwargs)
        self.innerclass = innerclass
        self.takes = takes
        self.returns = returns
        assert isinstance(body_attr, str)
        self.body_attr = body_attr
        if background_index is not None:
            self.background_indexref = _parse_index_like(background_index)
        else:
            self.background_indexref = None
        self.stemmer = stemmer
        self.stopwords = stopwords
        self.tokeniser = tokeniser
        self.kwargs = kwargs
        self.verbose = verbose

    def transform(self, topics_and_res):
        # we use _IterDictIndexer_nofifo, as _IterDictIndexer_fifo (which is default on unix) doesnt support IndexingType.MEMORY as a destination
        from pyterrier.terrier import IndexFactory
        from pyterrier.terrier.index import IndexingType, _IterDictIndexer_nofifo
        pt.validate.result_frame(topics_and_res, extra_columns=[self.body_attr, 'query'])
        documents = topics_and_res[["docno", self.body_attr]].drop_duplicates(subset="docno").rename(columns={self.body_attr:'text'})
        indexref = _IterDictIndexer_nofifo(
            None, 
            type=IndexingType.MEMORY, 
            verbose=self.verbose, 
            stemmer = self.stemmer, 
            stopwords = self.stopwords,
            tokeniser = self.tokeniser
            ).index(documents.to_dict(orient='records'))
        
        docno2docid = { docno:id for id, docno in enumerate(documents["docno"]) }
        index_docs = IndexFactory.of(indexref)
        docno2docid = {}
        for i in range(0, index_docs.getCollectionStatistics().getNumberOfDocuments()):
            docno2docid[index_docs.getMetaIndex().getItem("docno", i)] = i
        assert len(docno2docid) == index_docs.getCollectionStatistics().getNumberOfDocuments(), "docno2docid size (%d) doesnt match index (%d)" % (len(docno2docid), index_docs.getCollectionStatistics().getNumberOfDocuments())
        
        # if a background index is set, we create an "IndexWithBackground" using both that and our new index
        if self.background_indexref is None:
            index = index_docs
        else:
            index_background = IndexFactory.of(self.background_indexref)
            index = pt.java.autoclass("org.terrier.python.IndexWithBackground")(index_docs, index_background)          

        topics = topics_and_res[["qid", "query"]].dropna(axis=0, subset=["query"]).drop_duplicates()
        
        if self.takes == "queries":
            # we have provided the documents, so we dont need a docno or docid column that will confuse 
            # BR and think it is re-ranking. In fact, we only need qid and query
            input = topics
        elif self.takes == "docs":
            # we have to pass the documents, but its desirable to have the docids mapped to the new index already
            # build a mapping, as the metaindex may not have reverse lookups enabled
            input = topics_and_res.copy()
            # add the docid to the dataframe
            input["docid"] = input.apply(lambda row: docno2docid[row["docno"]], axis=1, result_type='reduce')

        if self.innerclass == Retriever:
            # build up a termpipeline based on the stemmer and stopwords settings
            tp = []
            stops, _ = TerrierStopwords._to_obj(self.stopwords or 'none')
            if stops == TerrierStopwords.terrier:
                tp.append("Stopwords")
            elif stops == TerrierStopwords.none:
                pass # noop
            else:
                # sadly PyTerrierCustomStopwordList only support instantiation from ApplicationSetup or Index properties, not controls
                raise KeyError("Only TerrierStopwords.terrier and TerrierStopwords.none are supported for stopwords in TextIndexProcessor, found %s" % str(stops))       
            stemmer = TerrierStemmer._to_obj(self.stemmer or 'none')
            if stemmer is not TerrierStemmer.none:
                tp.append(TerrierStemmer._to_class(stemmer))

            if 'controls' not in self.kwargs:
                self.kwargs['controls'] = {}
            if not len(tp):
                tp = ["NoOp"] # an empty termpipeline will be detected as missing by ApplyTermPipeline in Terrier
            self.kwargs['controls']['termpipelines'] = ",".join(tp)
        else:
            # if not Retriever, we dont know how to set the termpipeline
            pass

        # and then just instantiate BR using the our new index 
        # we take all other arguments as arguments for BR
        inner = self.innerclass(index, **(self.kwargs))
        inner.verbose = self.verbose
        inner_res = inner.transform(input)

        if self.returns == "docs":
            # as this is a new index, docids are not meaningful externally, so lets drop them
            inner_res.drop(columns=['docid'], inplace=True)

            topics_columns = topics_and_res.columns[(topics_and_res.columns.isin(["qid", "docno"])) | (~topics_and_res.columns.isin(inner_res.columns))]
            if len(inner_res) < len(topics_and_res):
                inner_res = topics_and_res[topics_columns].merge(inner_res, on=["qid", "docno"], how="left")
                inner_res["score"] = inner_res["score"].fillna(value=0)
            else:
                inner_res = topics_and_res[ topics_columns ].merge(inner_res, on=["qid", "docno"])
        elif self.returns == "queries":
            if len(inner_res) < len(topics):
                inner_res = topics.merge(on=["qid"], how="left")
        else:
            raise ValueError("returns attribute should be docs of queries")
        return inner_res


class TextScorer(TextIndexProcessor):
    """
        A re-ranker class, which takes the queries and the contents of documents, indexes the contents of the documents using a MemoryIndex, and performs ranking of those documents with respect to the queries.
        Unknown kwargs are passed to Retriever.

        :param takes: configuration - what is needed as input: `"queries"`, or `"docs"`. Default is `"docs"` since v0.8.
        :param returns: configuration - what is needed as output: `"queries"`, or `"docs"`. Default is `"docs"`.
        :param body_attr: what dataframe input column contains the text of the document. Default is `"body"`.
        :param wmodel: name of the weighting model to use for scoring.
        :param background_index: An optional background index to use for term and collection statistics. If a weighting
            model such as BM25 or TF_IDF or PL2 is used without setting the background_index, the background statistics
            will be calculated from the dataframe, which is ususally not the desired behaviour.

        Example::

            df = pd.DataFrame(
                [
                    ["q1", "chemical reactions", "d1", "professor protor poured the chemicals"],
                    ["q1", "chemical reactions", "d2", "chemical brothers turned up the beats"],
                ], columns=["qid", "query", "text"])
            textscorer = pt.TextScorer(takes="docs", body_attr="text", wmodel="Tf")
            rtr = textscorer.transform(df)
            #rtr will score each document by term frequency for the query "chemical reactions" based on the provided document contents

        Example::

            df = pd.DataFrame(
                [
                    ["q1", "chemical reactions", "d1", "professor protor poured the chemicals"],
                    ["q1", "chemical reactions", "d2", "chemical brothers turned up the beats"],
                ], columns=["qid", "query", "text"])
            existing_index = pt.IndexFactory.of(...)
            textscorer = pt.TextScorer(takes="docs", body_attr="text", wmodel="TF_IDF", background_index=existing_index)
            rtr = textscorer.transform(df)
            #rtr will score each document by TF_IDF for the query "chemical reactions" based on the provided document contents
    """

    def __init__(self, takes="docs", **kwargs):
        super().__init__(Retriever, takes=takes, **kwargs)


@pt.java.required
class FeaturesRetriever(Retriever):
    """
    Use this class for retrieval with multiple features
    """

    #: FBR_default_controls(dict): stores the default properties for a FBR
    FBR_default_controls = Retriever.default_controls.copy()
    FBR_default_controls["matching"] = "FatFeaturedScoringMatching,org.terrier.matching.daat.FatFull"
    del FBR_default_controls["wmodel"]
    #: FBR_default_properties(dict): stores the default properties
    FBR_default_properties = Retriever.default_properties.copy()

    def __init__(self, 
                 index_location : Union[str,Any], 
                 features : List[str], 
                 controls : Optional[Dict[str,str]] = None, 
                 properties : Optional[Dict[str,str]] = None, 
                 threads : int = 1, 
                 **kwargs):
        """
            Init method

            :param index_location: An index-like object - An Index, an IndexRef, or a String that can be resolved to an IndexRef
            :param features: List of features to use
            :param controls: A dictionary with the control names and values
            :param properties: A dictionary with the property keys and values
            :param verbose: If True transform method will display progress
            :param num_results: Number of results to retrieve. 
        """
        controls = _mergeDicts(FeaturesRetriever.FBR_default_controls, controls)
        properties = _mergeDicts(FeaturesRetriever.FBR_default_properties, properties)
        self.features = features
        properties["fat.featured.scoring.matching.features"] = ";".join(features)

        # record the weighting model
        self.wmodel = None
        if "wmodel" in kwargs and kwargs['wmodel'] is not None:
            assert isinstance(kwargs["wmodel"], str), "Non-string weighting models not yet supported by FBR"
            self.wmodel = kwargs["wmodel"]
        if "wmodel" in controls:
            self.wmodel = controls["wmodel"]
        
        # check for terrier-core#246 bug usiung FatFull
        if self.wmodel is not None:    
            assert pt.terrier.check_version(5.9), "Terrier 5.9 is required for this functionality, see https://github.com/terrier-org/terrier-core/pull/246"
            
        if threads > 1:
            raise ValueError("Multi-threaded retrieval not yet supported by FeaturesRetriever")
        
        super().__init__(index_location, controls, properties, **kwargs)
        if self.wmodel is None and 'wmodel' in self.controls:
            del self.controls['wmodel'] # Retriever sets a default controls['wmodel'], we only want this

    def __reduce__(self):
        return (
            self.__class__,
            (self.indexref, self.features),
            self.__getstate__()
        )
    
    def schematic(self, *, input_columns = None): 
        if self.wmodel is None:
            return {'label': "FeaturesRetriever: %df" % len(self.features)}
        return {'label': "FeaturesRetriever: %s + %df" % (self.wmodel, len(self.features))}

    def __getstate__(self): 
        return  {
                'controls' : self.controls, 
                'properties' : self.properties, 
                'metadata' : self.metadata,
                'features' : self.features,
                'wmodel' : self.wmodel
                #TODO consider the context state?
                }

    def __setstate__(self, d): 
        self.controls = d["controls"]
        self.metadata = d["metadata"]
        self.features = d["features"]
        self.wmodel = d["wmodel"]
        self.properties.update(d["properties"])
        for key,value in d["properties"].items():
            pt.terrier.J.ApplicationSetup.setProperty(key, str(value))
        #TODO consider the context state?

    @staticmethod 
    def from_dataset(dataset : Union[str,Dataset], 
            variant : Optional[str] = None, 
            version='latest',            
            **kwargs):
        return pt.datasets.transformer_from_dataset(dataset, variant=variant, version=version, clz=FeaturesRetriever, **kwargs)

    def transform(self, queries):
        """
        Performs the retrieval with multiple features

        :param queries: A pandas.Dataframe with columns=['qid', 'query']. For re-ranking,
                the DataFrame may also have a 'docid' and or 'docno' column.

        :return: a pandas.DataFrame with columns=['qid', 'docno', 'score', 'rank, 'features']
        """
        results = []
        if not isinstance(queries, pd.DataFrame):
            raise ValueError(".transform() should be passed a dataframe. Use .search() to execute a single query; Use .transform_iter() for iter-dicts")

        # use pt.validate - this makes inspection of input columns better
        with pt.validate.any(queries) as v:
            v.columns(includes=['qid', 'query', 'docid'], mode='rerank') # docid-based results frame
            v.query_frame(extra_columns=['query_toks'], mode='retrieve_toks')
            v.query_frame(extra_columns=['query'], mode='retrieve')
            v.result_frame(extra_columns=['query'], mode='rerank')

        docno_provided = "docno" in queries.columns
        docid_provided = "docid" in queries.columns
        scores_provided = "score" in queries.columns
        if v.mode == 'rerank':
            assert docno_provided or docid_provided, "For reranking, either docno or docid must be provided"
            #re-ranking mode
            assert pt.terrier.check_version(5.3)
            input_results = queries

            # query is optional, and functionally dependent on qid.
            # Hence as long as one row has the query for each qid, 
            # the rest can be None
            queries = input_results[["qid", "query"]].dropna(axis=0, subset=["query"]).drop_duplicates()
            RequestContextMatching = pt.java.autoclass("org.terrier.python.RequestContextMatching")

            if not scores_provided and self.wmodel is None:
                raise ValueError("We're in re-ranking mode, but input does not have scores, and wmodel is None")
        else:
            assert v.mode == 'retrieve' or v.mode == 'retrieve_toks'
            assert not scores_provided

            if self.wmodel is None:
                raise ValueError("We're in retrieval mode (input columns were "+str(queries.columns)+"), but wmodel is None. FeaturesRetriever requires a wmodel be set for identifying the candidate set. "
                    +" Hint: wmodel argument for FeaturesRetriever, e.g. FeaturesRetriever(index, features, wmodel=\"DPH\")")

        if queries["qid"].dtype == np.int64:
            queries['qid'] = queries['qid'].astype(str)

        newscores=[]
        iter = queries.itertuples()
        if self.verbose and len(queries):
            iter = pt.tqdm(iter, desc=str(self), total=queries.shape[0], unit="q")
        for row in iter:
            qid = str(row.qid)
            query_toks_present = 'query_toks' in row._fields
            if query_toks_present:
                # Even though it might look like we should parse the query toks here, we don't want the resulting query to be caught by the conditions
                # that come before the "if query_toks_present" check. So we set it to an empty string and handle the parsing below.
                query = ''
                srq = self.manager.newSearchRequest(qid)
            else:
                query = row.query
                if _query_needs_tokenised(query):
                    query = ' '.join(self.tokeniser.getTokens(query))
                if len(query) == 0:
                    warn(
                        "Skipping empty query for qid %s" % qid)
                    continue
                srq = self.manager.newSearchRequest(qid, query)

            for control, value in self.controls.items():
                srq.setControl(control, str(value))

            # this is needed until terrier-core issue #106 lands
            if "applypipeline:off" in query:
                srq.setControl("applypipeline", "off")
                srq.setOriginalQuery(query.replace("applypipeline:off", ""))

            # transparently detect matchop queries
            if _matchop(query):
                srq.setControl("terrierql", "off")
                srq.setControl("parsecontrols", "off")
                srq.setControl("parseql", "off")
                srq.setControl("matchopql", "on")

            if query_toks_present:
                if len(row.query_toks) == 0:
                    warn(
                        "Skipping empty query_toks for qid %s" % qid)
                    return []
                srq.setControl("terrierql", "off")
                srq.setControl("parsecontrols", "off")
                srq.setControl("parseql", "off")
                srq.setControl("matchopql", "on")
                query = _querytoks2matchop(row.query_toks)
                srq.setOriginalQuery(query)

            # this handles the case that a candidate set of documents has been set. 
            if docno_provided or docid_provided:
                # we use RequestContextMatching to make a ResultSet from the 
                # documents in the candidate set. 
                matching_config_factory = RequestContextMatching.of(srq)
                input_query_results = input_results[input_results["qid"] == qid]
                if docid_provided:
                    matching_config_factory.fromDocids(input_query_results["docid"].values.tolist())
                elif docno_provided:
                    matching_config_factory.fromDocnos(input_query_results["docno"].values.tolist())
                if scores_provided:
                    if self.wmodel is None:
                        # we provide the scores, so dont use a weighting model, and pass the scores through Terrier
                        matching_config_factory.withScores(input_query_results["score"].values.tolist())
                        srq.setControl("wmodel", "Null")
                    else:
                        srq.setControl("wmodel", self.wmodel)
                matching_config_factory.build()
                srq.setControl("matching", ",".join(["FatFeaturedScoringMatching","ScoringMatchingWithFat", srq.getControl("matching")]))
            
            self.manager.runSearchRequest(srq)
            srq = pt.java.cast('org.terrier.querying.Request', srq)
            fres = pt.java.cast('org.terrier.learning.FeaturedResultSet', srq.getResultSet())
            feat_names = fres.getFeatureNames()

            docids=fres.getDocids()
            scores= fres.getScores()
            metadata_list = [fres.getMetaItems(meta_column) for meta_column in self.metadata]
            feats_values = [fres.getFeatureScores(feat) for feat in feat_names]
            rank = FIRST_RANK
            for i in range(fres.getResultSize()):
                doc_features = np.array([ feature[i] for feature in feats_values])
                meta=[ metadata_col[i] for metadata_col in metadata_list]
                results.append( [qid, query, docids[i], rank, doc_features ] + meta )
                newscores.append(scores[i])
                rank += 1

        res_dt = pd.DataFrame(results, columns=["qid", "query", "docid", "rank", "features"] + self.metadata)
        res_dt["score"] = newscores
        # ensure to return the query and any other input columns
        input_cols = queries.columns[ (queries.columns == "qid") | (~queries.columns.isin(res_dt.columns))]
        res_dt = res_dt.merge(queries[input_cols], on=["qid"])
        return res_dt

    def __repr__(self):
        return "TerrierFeatRetr(" + ",".join([
            self.indexref.toString(),
            str(self.features),
            str(self.controls),
            str(self.properties)
        ]) + ")"

    def __str__(self):
        if self.wmodel is None:
            return "TerrierFeatRetr(" + str(len(self.features)) + " features)"
        return "TerrierFeatRetr(" + self.controls["wmodel"] + " and " + str(len(self.features)) + " features)"

    def fuse_left(self, left: pt.Transformer) -> Optional[pt.Transformer]:
        # Can merge Retriever >> FeaturesRetriever into a single FeaturesRetriever that also retrieves
        # if the indexref matches and the current FeaturesRetriever isn't already reranking.
        if isinstance(left, Retriever) and \
           self.indexref == left.indexref and \
           left.controls.get('context_wmodel') != 'on' and \
           self.wmodel is None:
            return FeaturesRetriever(
                self.indexref,
                self.features,
                controls=self.controls,
                properties=self.properties,
                threads=self.threads,
                wmodel=left.controls['wmodel'],
            )
        return None

    def fuse_rank_cutoff(self, k: int) -> Optional[pt.Transformer]:
        """
        Support fusing with RankCutoffTransformer.
        """
        if float(self.controls.get('end', float('inf'))) < k:
            return self # the applied rank cutoff is greater than the one already applied
        if self.wmodel is None:
            return None # not a retriever
        # apply the new k as num_results
        return FeaturesRetriever(self.indexref, self.features, controls=self.controls, properties=self.properties,
            threads=self.threads, wmodel=self.wmodel, verbose=self.verbose, num_results=k)

    def fuse_feature_union(self, other: pt.Transformer, is_left: bool) -> Optional[pt.Transformer]:
        if isinstance(other, FeaturesRetriever) and \
           self.indexref == other.indexref and \
           self.wmodel is None and \
           other.wmodel is None:
            features = self.features + other.features if is_left else other.features + self.features
            return FeaturesRetriever(self.indexref, features, controls=self.controls, properties=self.properties,
                threads=self.threads, wmodel=self.wmodel, verbose=self.verbose)

        if isinstance(other, Retriever) and \
           self.indexref == other.indexref and \
           self.wmodel is None  and \
           other.controls.get('context_wmodel') != 'on':
            features = self.features + ["WMODEL:" + other.controls['wmodel']] if is_left else ["WMODEL:" + other.controls['wmodel']] + self.features
            return FeaturesRetriever(self.indexref, features, controls=self.controls, properties=self.properties,
                threads=self.threads, wmodel=self.wmodel, verbose=self.verbose)
        
        return None




from .base import TempDirTestCase
import warnings
import tempfile
import shutil
import os
import pyterrier as pt

class TestFilesIndexer(TempDirTestCase):

    def test_2_docs(self):
        files = pt.io.find_files(os.path.join(self.here, "fixtures", "sample_docs"))
        indexer = pt.FilesIndexer(self.test_dir)
        indexref = indexer.index(files)
        index = pt.IndexFactory.of(indexref)
        self.assertEqual(2, index.getCollectionStatistics().getNumberOfDocuments())

    def test_2_docs_custom__stops(self):
        files = pt.io.find_files(os.path.join(self.here, "fixtures", "sample_docs"))
        indexer = pt.FilesIndexer(self.test_dir, stopwords=['empty'])
        indexref = indexer.index(files)
        index = pt.IndexFactory.of(indexref)
        self.assertEqual(2, index.getCollectionStatistics().getNumberOfDocuments())
        self.assertFalse("empty" in index.getLexicon())

    def test_2_docs_title_body_meta(self):
        sample_dir = os.path.join(self.here, "fixtures", "sample_docs")
        files = pt.io.find_files(sample_dir)
        indexer = pt.FilesIndexer(
            self.test_dir, 
            meta={"docno" : 20, "filename" : 512, "title" : 20, "body" : 40}, 
            meta_tags={"title":"title", "body" : "ELSE"})
        indexref = indexer.index(files)
        index = pt.IndexFactory.of(indexref)
        
        # check index size
        self.assertEqual(2, index.getCollectionStatistics().getNumberOfDocuments())
        print(index.getMetaIndex().getAllItems(0))
        print(index.getMetaIndex().getAllItems(1))

        # determine file locations as docids
        html_file = os.path.join(sample_dir, "a.html")
        html_pos = files.index(html_file)
        txt_file = os.path.join(sample_dir, "b.txt")
        txt_pos = files.index(txt_file)        

        self.assertTrue(html_pos < len(files))
        # test filename -> docid lookup
        self.assertEqual(html_pos, index.getMetaIndex().getDocument("filename", html_file))
        # test docid -> filename lookup
        self.assertEqual(html_file, index.getMetaIndex().getItem("filename", html_pos))
        # test title has been recorded in metaindex
        self.assertEqual("test title", index.getMetaIndex().getItem("title", html_pos))

        if not pt.terrier.check_version("5.5"):
            return

        # test bodies have been recorded in metaindex
        self.assertEqual("test body", index.getMetaIndex().getItem("body", html_pos))
        self.assertEqual("empty text document",index.getMetaIndex().getItem("body", txt_pos))



from typing import Sequence, Union, Optional, cast, Iterable
import pandas as pd
from .model import add_ranks

def empty_Q() -> pd.DataFrame:
    """
        Returns an empty dataframe with columns `["qid", "query"]`.
    """
    return pd.DataFrame(columns=["qid", "query"])

def queries(queries : Union[str, Sequence[str]], qid : Optional[Union[str, Iterable[str]]] = None, **others) -> pd.DataFrame:
    """
        Creates a new queries dataframe. Will return a dataframe with the columns `["qid", "query"]`. 
        Any further lists in others will also be added.

        Arguments:
            queries: The search queries. Either a string, for a single query, or a sequence (e.g. list of strings)
            qids: Corresponding query ids. Either a string, for a single query, or a sequence (e.g. list of strings). Must have same length as queries.
            others: A dictionary of other attributes to add to the query dataframe           

        Examples::

            # create a dataframe with one query, qid "1"
            one_query = pt.new.queries("what the noise was was the question")

            # create a dataframe with one query, qid "5"
            one_query = pt.new.queries("what the noise was was the question", 5)

            # create a dataframe with two queries
            one_query = pt.new.queries(["query text A", "query text B"], ["1", "2"])

            # create a dataframe with two queries
            one_query = pt.new.queries(["query text A", "query text B"], ["1", "2"], categories=["catA", "catB"])
  
    """
    if isinstance(queries, str):
        if qid is None:
            qid = "1"
        assert isinstance(qid, str)
        return pd.DataFrame({"qid" : [qid], "query" : [queries], **others})
    if qid is None:
        qid = cast(Iterable[str], map(str, range(1, len(queries)+1))) # noqa: PT100 (this is typing.cast, not jinus.cast)
    return pd.DataFrame({"qid" : qid, "query" : queries, **others})

Q = queries

def empty_R() -> pd.DataFrame:
    """
        Returns an empty dataframe with columns `["qid", "query", "docno", "rank", "score"]`.
    """
    return pd.DataFrame([[]], columns=["qid", "query", "docno", "rank", "score"])

def ranked_documents(
        scores : Sequence[Sequence[float]], 
        qid : Optional[Sequence[str]] = None, 
        docno : Optional[Sequence[Sequence[str]]] = None, 
        **others) -> pd.DataFrame:
    """
        Creates a new ranked documents dataframe. Will return a dataframe with the columns `["qid", "docno", "score", "rank"]`. 
        Any further lists in others will also be added.

        Arguments:
            scores: The scores of the retrieved documents. Must be a list of lists.
            qid: Corresponding query ids. Must have same length as the first dimension of scores.
                If omitted, documents, qids are computed as strings starting from "1"
            docno: Corresponding docnos.  Must have same length as the first dimension of scores 
                and each 2nd dimension must be the same as the number of documents retrieved.
                If omitted, docnos are computed as strings starting from "d1" for each query.
            others: A dictionary of other attributes to add to the query dataframe.         

        Examples::

            # one query, one document
            R1 = pt.new.ranked_documents([[1]])

            # one query, two documents
            R2 = pt.new.ranked_documents([[1, 2]])

            # two queries, one documents each
            R3 = pt.new.ranked_documents([[1], [2]])

            # one query, one document, qid specified
            R4 = pt.new.ranked_documents([[1]], qid=["q100"])

            # one query, one document, qid and docno specified
            R5 = pt.new.ranked_documents([[1]], qid=["q100"], docno=[["d20"]])

    """
    from itertools import chain
    if len(scores) == 0:
        return empty_R()
    rtr = None
    if isinstance(scores[0], list):
        # multiple queries
        if qid is None:
            qid = list(map(str, range(1, len(scores)+1)))
        else:
            assert len(qid) == len(scores)
        qid = list(chain.from_iterable([ [q] * len(score_array) for q, score_array in zip(qid, scores) ]))
        
        if docno is None:
            docno = [ list(map(lambda i: "d%d" % i, range(1, len(score_array)+1) ) ) for score_array in scores ]
        else:
            assert len(docno) == len(scores)
        
        rtr = pd.DataFrame(list(chain.from_iterable(scores)), columns=["score"]) 
        
        rtr["docno"] = list(chain.from_iterable(docno))
        rtr["qid"] = qid
        #construct = {"qid" : qid, "docno" : docno, "score" : scores}
        for k, v in others.items():
            rtr[k] = list(chain.from_iterable(v))
            #assert len(v) == len(scores), "kwarg %s had length %d but was expected to have length %d" % (k, len(v), len(scores))
            #construct[k] = np.array( v ).flatten()
        #rtr = pd.DataFrame(construct)        
    else:
        raise ValueError("We assume multiple documents, for now")
    return add_ranks(rtr)

R = ranked_documents



import pandas as pd
import pyterrier as pt
import os
import unittest
import warnings
import math
from pyterrier.measures import *
from .base import TempDirTestCase
import pytest

class TestExperiment(TempDirTestCase):

    def __init__(self, *args, **kwargs):
        super(TempDirTestCase, self).__init__(*args, **kwargs)
        self.index = None

    def _vaswani_index(self):
        if self.index is not None:
            return self.index
        indexloc = self.here + "/fixtures/index/data.properties"
        self.index = pt.IndexFactory.of(indexloc)
        return self.index
    
    def test_experiment_render(self):
        from pyterrier._evaluation._rendering import RenderFromPerQuery
        r = RenderFromPerQuery(['bm25'])
        r.add_metrics(0, {'q1': {'AP' : 1.0}, 'q2': {'AP' : 0.8}}, 1000)
        df = r.perquery()
        self.assertEqual(2, len(df))
        self.assertEqual(1.8, df['value'].sum())
        df = r.averages()
        self.assertEqual(1, len(df))
        self.assertEqual(0.9, df['AP'].iloc[0])


    def test_precomp_common(self):
        bm25 = pt.terrier.Retriever(self._vaswani_index(), wmodel='BM25')
        pipeA = bm25 %3
        pipeB = bm25 %10
        import pyterrier._evaluation._execution
        common, suffices = pyterrier._evaluation._execution._identifyCommon([pipeA, pipeB])
        self.assertEqual(bm25, common)
        self.assertIsInstance(suffices[0], pt.RankCutoff)
        self.assertEqual(3, suffices[0].k)
        self.assertIsInstance(suffices[1], pt.RankCutoff)
        self.assertEqual(10, suffices[1].k)

        common, suffices = pyterrier._evaluation._execution._identifyCommon([bm25, pipeB])
        self.assertEqual(bm25, common)
        self.assertIsInstance(suffices[0], pt.transformer.IdentityTransformer)
        self.assertIsInstance(suffices[1], pt.RankCutoff)
        self.assertEqual(10, suffices[1].k)
    
    def test_precompute_experiment(self):
        bm25 = pt.terrier.Retriever(self._vaswani_index(), wmodel='BM25')
        pipeB = bm25 %10
        df1 = pt.Experiment([bm25, pipeB], pt.get_dataset('vaswani').get_topics().head(10), pt.get_dataset('vaswani').get_qrels(), eval_metrics=['map'])
        df2 = pt.Experiment([bm25, pipeB], pt.get_dataset('vaswani').get_topics().head(10), pt.get_dataset('vaswani').get_qrels(), eval_metrics=['map'], precompute_prefix=True)
        pd.testing.assert_frame_equal(df1, df2)

        df3 = pt.Experiment([bm25, pipeB], pt.get_dataset('vaswani').get_topics().head(10), pt.get_dataset('vaswani').get_qrels(), eval_metrics=['map'], precompute_prefix=True, batch_size=4)
        pd.testing.assert_frame_equal(df1, df3)

    def test_irm_APrel2(self):
        topics = pd.DataFrame([["q1", "q1"], ["q2", "q1"] ], columns=["qid", "query"])
        res1 = pd.DataFrame([["q1", "d1", 1.0], ["q2", "d1", 2.0] ], columns=["qid", "docno", "score"])
        qrels = pd.DataFrame([["q1", "d1", 1], ["q2", "d1", 2] ], columns=["qid", "docno", "label"])
        df = pt.Experiment(
                [res1],
                topics,
                qrels,
                [
                    pt.measures.AP(rel=2),
                ])
        self.assertEqual(0.5, df.iloc[0]["AP(rel=2)"])
        df = pt.Experiment(
                [res1],
                topics,
                qrels,
                [
                    pt.measures.AP,
                    pt.measures.AP(rel=2),
                    pt.measures.P(rel=2)@1
                ])
        print(df.columns)
        self.assertEqual(1, df.iloc[0]["AP"])
        self.assertEqual(0.5, df.iloc[0]["AP(rel=2)"])
        self.assertEqual(0.5, df.iloc[0]["P(rel=2)@1"])
        df = pt.Experiment(
                [res1],
                topics,
                qrels,
                [
                    pt.measures.AP,
                    pt.measures.AP(rel=2),
                    pt.measures.P(rel=2)@1
                ],
                perquery=True)
        self.assertEqual(1, df[(df.measure == "AP") & (df.qid == "q1")].value.iloc[0])
        self.assertEqual(1, df[(df.measure == "AP") & (df.qid == "q2")].value.iloc[0])
        self.assertEqual(0, df[(df.measure == "AP(rel=2)") & (df.qid == "q1")].value.iloc[0])
        self.assertEqual(1, df[(df.measure == "AP(rel=2)") & (df.qid == "q2")].value.iloc[0])
        self.assertEqual(0, df[(df.measure == "P(rel=2)@1") & (df.qid == "q1")].value.iloc[0])
        self.assertEqual(1, df[(df.measure == "P(rel=2)@1") & (df.qid == "q2")].value.iloc[0])

    def test_differing_queries(self):
        bm25 = pt.terrier.Retriever(self._vaswani_index(), wmodel='BM25')
        qrels = pd.DataFrame({
            'qid':   ["1",     "1",    "2"],
            'docno': ["10703", "1056", "9374"],
            'label': [1,       1,      1],
        })
        topics = pd.DataFrame({
            'qid':   ["1",        "2",         "3"],
            'query': ["chemical", "reactions", "reaction"]
        })
        test_cases = [
            # perfect topic/qrel overlap, filter_by_qrels and filter_by_topics have no effect
            (['1', '2'], ['1', '2'], True,  True,   {'1': 1.0, '2': 0.5}),
            (['1', '2'], ['1', '2'], True,  False,  {'1': 1.0, '2': 0.5}),
            (['1', '2'], ['1', '2'], False, True,   {'1': 1.0, '2': 0.5}),
            (['1', '2'], ['1', '2'], False, False,  {'1': 1.0, '2': 0.5}),
            # qid=2 missing from topics; qid=2 should only be included if filter_by_topics=False, filter_by_qrels has no effect
            (['1', '2'], ['1'], True,  True,   {'1': 1.0}),
            (['1', '2'], ['1'], True,  False,  {'1': 1.0, '2': 0.}),
            (['1', '2'], ['1'], False, True,   {'1': 1.0}),
            (['1', '2'], ['1'], False, False,  {'1': 1.0, '2': 0.}),
            # qid=2 missing from qrels; qid=2 should never be included in the results, '2' should be NaN if filter_by_qrels=False
            (['1'], ['1', '2'], True,  True,   {'1': 1.0}),
            (['1'], ['1', '2'], True,  False,  {'1': 1.0}),
            (['1'], ['1', '2'], False, True,   {'1': 1.0, '2': float('NaN')}),
            (['1'], ['1', '2'], False, False,  {'1': 1.0, '2': float('NaN')}),
            # qid=3 missing from qrels and qid=1 is missing from the topics; qid=1 should only be included if filter_by_topics=False
            (['1', '2'], ['2', '3'], True,  True,   {'2': 0.5}),
            (['1', '2'], ['2', '3'], True,  False,  {'1': 0.0, '2': 0.5}),
            (['1', '2'], ['2', '3'], False, True,   {'2': 0.5, '3': float('NaN')}),
            (['1', '2'], ['2', '3'], False, False,  {'1': 0.0, '2': 0.5, '3': float('NaN')}),
            # no qid overlap between topics and qrels; should throw exception if filter_by_topics=True
            (['1'], ['3'], True,  True,   ValueError('There is no overlap between the qids found in the topics and qrels. If this is intentional, set filter_by_topics=False and filter_by_qrels=False.')),
            (['1'], ['3'], True,  False,  ValueError('There is no overlap between the qids found in the topics and qrels. If this is intentional, set filter_by_topics=False and filter_by_qrels=False.')),
            (['1'], ['3'], False, True,   ValueError('There is no overlap between the qids found in the topics and qrels. If this is intentional, set filter_by_topics=False and filter_by_qrels=False.')),
            (['1'], ['3'], False, False,  {'1': 0.0, '3': float('NaN')}),
        ]
        for qrel_qids, topic_qids, filter_by_qrels, filter_by_topics, result in test_cases:
            for batch_size in [1]:
                with self.subTest(f'qrel_qids={qrel_qids} topic_qids={topic_qids} filter_by_qrels={filter_by_qrels} filter_by_topics={filter_by_topics} batch_size={batch_size}'):
                    if isinstance(result, ValueError):
                        with self.assertRaises(ValueError) as context:
                            pt.Experiment([bm25], topics[topics.qid.isin(topic_qids)], qrels[qrels.qid.isin(qrel_qids)], [P@2, 'P', 'mrt'], filter_by_qrels=filter_by_qrels, filter_by_topics=filter_by_topics, perquery=True, batch_size=batch_size)
                        self.assertEqual(context.exception.args, result.args)
                        with self.assertRaises(ValueError) as context:
                            pt.Experiment([bm25], topics[topics.qid.isin(topic_qids)], qrels[qrels.qid.isin(qrel_qids)], [P@2, 'P', 'mrt'], filter_by_qrels=filter_by_qrels, filter_by_topics=filter_by_topics, perquery=False, batch_size=batch_size)
                        self.assertEqual(context.exception.args, result.args)
                    else:
                        with warnings.catch_warnings(record=True) as w:
                            res = pt.Experiment([bm25], topics[topics.qid.isin(topic_qids)], qrels[qrels.qid.isin(qrel_qids)], [P@2, 'P', 'mrt'], filter_by_qrels=filter_by_qrels, filter_by_topics=filter_by_topics, perquery=True, batch_size=batch_size)
                        if any(math.isnan(v) for v in result.values()):
                            self.assertEqual(len(w), 1)
                            self.assertEqual(w[0].message.args[0], f'1 topic(s) not found in qrels. Scores for these topics are given as NaN and should not contribute to averages.')
                        else:
                            self.assertEqual(len(w), 0)
                        res = res[res['measure'] == 'P@2'].drop(columns=['name', 'measure'])
                        expected_res = pd.DataFrame([{'qid': qid, 'value': val} for qid, val in result.items()])
                        pd.testing.assert_frame_equal(res.reset_index(drop=True), expected_res.reset_index(drop=True))
                        res = pt.Experiment([bm25], topics[topics.qid.isin(topic_qids)], qrels[qrels.qid.isin(qrel_qids)], [P@2, 'P', 'mrt'], filter_by_qrels=filter_by_qrels, filter_by_topics=filter_by_topics, perquery=False, batch_size=batch_size)
                        num_result = {k: v for k, v in result.items() if not math.isnan(v)}
                        self.assertEqual(res.loc[0, 'P@2'], sum(num_result.values())/len(num_result))

    def test_wrong(self):
        brs = [
            pt.terrier.Retriever(self._vaswani_index(), wmodel='DPH'),
            pt.terrier.Retriever(self._vaswani_index(), wmodel='BM25')
        ]
        topics = pt.datasets.get_dataset("vaswani").get_topics().head(10)
        qrels =  pt.datasets.get_dataset("vaswani").get_qrels()
        with self.assertRaises(TypeError):
            pt.Experiment(brs, topics, qrels, eval_metrics=["map"], filter_qrels=True)
        
    def test_expected_columns(self):
        br0 = pt.Transformer.from_df(pd.DataFrame([{'qid' : '1', 'query' : 'hello'}]), uniform=True)
        br1 = pt.Transformer.from_df(pd.DataFrame([{'qid' : '1', 'query' : 'hello', "context" : "here"}]), uniform=True)

        topics = pt.datasets.get_dataset("vaswani").get_topics().head(10)
        qrels =  pt.datasets.get_dataset("vaswani").get_qrels()
        # check that we have the expected columns
        with self.assertRaises(ValueError) as ve:
            pt.Experiment([br1], topics, qrels, eval_metrics=["map"], validate='error')
            self.assertIn("Transformer", str(ve.exception))
            self.assertIn("does not produce all required columns", str(ve.exception))

        # lets make a custom measure that requires the 'context' column
        import ir_measures
        custom_measure = ir_measures.define_byquery(lambda qrels, run: len(run.iloc[0]["context"]), run_inputs=['context'], name='context_length')
        with self.assertRaises(ValueError) as ve:
            pt.Experiment([br0], topics, qrels, eval_metrics=[custom_measure], validate='error')
            self.assertIn("Transformer", str(ve.exception))
            self.assertIn("does not produce all required columns", str(ve.exception))
        eval = pt.Experiment([br1], topics, qrels, eval_metrics=[custom_measure], validate='error')
        self.assertIn("context_length", eval.columns)
        self.assertEqual(eval.iloc[0]["context_length"], 0.4)  # "here" has length 4, divide by 10 topics in the qrels.

        # check that we can detect a renaming of a column that is missing
        # this should fail, as the 'context' column is not present in the output of the br0 transformer
        with self.assertRaises(ValueError) as ve:
            pt.Experiment([br0 >> pt.apply.rename({'context' : 'prompt'})], topics, qrels, eval_metrics=[custom_measure], validate='error')

        # this one doesnt set expected columns, and doesnt support the empty df trick.
        def _rename_context(df):
            if len(df) == 0:
                raise ValueError("Empty DataFrame")
            return df.rename(columns={'context' : 'prompt'})
        with self.assertRaises(KeyError) as ke:
            with warnings.catch_warnings(record=True) as w:
            # this should give a warning, as the 'context' column is not present in the output of the br0 transformer, but we cant validate the rename
            # the experiment will then fail with a KeyError, as the 'context' column is not present in the output of the br0 transformer
                pt.Experiment([br0 >> pt.apply.generic(_rename_context)], topics, qrels, eval_metrics=[custom_measure], validate='error')
                # context is not present in the output of the br0 transformer, so we should get a KeyError
                self.assertIn("context", str(ke.exception))
                # but we coudlnt validate the rename, so we should get a warning, even if validate='error'
                self.assertIn("at position 0 failed to validate: Cannot determine output", str(w))

        
    def test_mrt(self):
        index = self._vaswani_index()
        brs = [
            pt.terrier.Retriever(index, wmodel="DPH"), 
            pt.terrier.Retriever(index, wmodel="BM25")
        ]
        topics = pt.datasets.get_dataset("vaswani").get_topics().head(10)
        qrels =  pt.datasets.get_dataset("vaswani").get_qrels()
        measures = ["map", "mrt"]
        pt.Experiment(brs, topics, qrels, eval_metrics=measures)
        self.assertTrue("mrt" in measures)
        pt.Experiment(brs, topics, qrels, eval_metrics=["map", "mrt"], highlight="color")
        pt.Experiment(brs, topics, qrels, eval_metrics=["map", "mrt"], baseline=0, highlight="color")

    def test_save_invalid_format(self):
        topics = pt.datasets.get_dataset("vaswani").get_topics().head(10)
        qrels =  pt.datasets.get_dataset("vaswani").get_qrels()
        t = pt.Transformer.from_df(pd.DataFrame([{'qid' : topics.iloc[0].qid, 'query' : 'hello', "context" : "here"}]))
        with self.assertRaises(TypeError) as te:
            pt.Experiment([t], topics, qrels, ['map'], save_dir=self.test_dir, names=['t'])
            self.assertIn("missing ['docno', 'score', 'rank']" in te.msg)
        # save_format with pickle should work - we need a dummy measure though.
        dummy_measure = ir_measures.define_byquery(lambda a,b: 0)
        import pickle
        pt.Experiment([t], 
                      topics, 
                      qrels, 
                      [dummy_measure], 
                      save_dir=self.test_dir, 
                      names=['t'], 
                      save_format=pickle)

    def test_save(self):
        index = self._vaswani_index()
        brs = [
            pt.terrier.Retriever(index, wmodel="DPH"), 
            pt.terrier.Retriever(index, wmodel="BM25")
        ]
        topics = pt.datasets.get_dataset("vaswani").get_topics().head(10)
        qrels =  pt.datasets.get_dataset("vaswani").get_qrels()

        import pickle
        for name, format, ext in [
                ('trec', 'trec', 'res.gz'),
                ('pkl_manual', pickle, 'mod'),
                ('pandas', (pd.read_csv, pd.DataFrame.to_csv), 'custom')
            ]: 
            with self.subTest(name):
                df1 = pt.Experiment(brs, topics, qrels, eval_metrics=["map", "mrt"], save_dir=self.test_dir, save_format=format)
                print("\n". join(os.listdir(self.test_dir)))
                # check save_dir files are there
                self.assertTrue(os.path.exists(os.path.join(self.test_dir, "TerrierRetr(DPH)." + ext)), os.path.join(self.test_dir, "TerrierRetr(DPH)." + ext) + " not found")
                self.assertTrue(os.path.exists(os.path.join(self.test_dir, "TerrierRetr(BM25)." + ext)), os.path.join(self.test_dir, "TerrierRetr(BM25)." + ext) + " not found")

                # check for warning
                with pytest.warns(UserWarning):
                    # reuse only kicks in when save_mode is set.
                    df2 = pt.Experiment(brs, topics, qrels, eval_metrics=["map", "mrt"], save_dir=self.test_dir, save_format=format)

                # check for error when save_mode='error'
                with self.assertRaises(ValueError):
                    # reuse only kicks in when save_mode is set.
                    df2 = pt.Experiment(brs, topics, qrels, eval_metrics=["map", "mrt"], save_dir=self.test_dir, save_mode='error', save_format=format)

                # allow it to reuse
                df2 = pt.Experiment(brs, topics, qrels, eval_metrics=["map", "mrt"], save_dir=self.test_dir, save_mode='reuse', save_format=format)
                # a successful experiment using save_dir should be faster
                self.assertTrue(df2.iloc[0]["mrt"] < df1.iloc[0]["mrt"])
        
    def test_empty(self):
        df1 = pt.new.ranked_documents([[1]]).head(0)
        t1 = pt.Transformer.from_df(df1)

        topics = pt.datasets.get_dataset("vaswani").get_topics().head(10)
        qrels =  pt.datasets.get_dataset("vaswani").get_qrels()
        with self.assertRaises(ValueError):
            pt.Experiment([df1], topics, qrels, eval_metrics=["map", "mrt"])
        with self.assertRaises(ValueError):
            pt.Experiment([t1], topics, qrels, eval_metrics=["map", "mrt"])
        with self.assertRaises(ValueError):
            pt.Experiment([t1], topics, qrels, eval_metrics=["map", "mrt"], batch_size=2)
        

    def test_various_metrics(self):
        topics = pt.datasets.get_dataset("vaswani").get_topics().head(10)
        index = self._vaswani_index()
        res = [
            pt.terrier.Retriever(index, wmodel="DPH")(topics), 
            pt.terrier.Retriever(index, wmodel="BM25")(topics)
        ]
        
        qrels =  pt.datasets.get_dataset("vaswani").get_qrels()
        # what we ask for -> what we should get as a metric
        family2measure = {
            'ndcg_cut_5' : 'ndcg_cut_5',
            'P' : "P@5",
            'P_5' : "P_5",
            "iprec_at_recall" : "IPrec@0.0",
            "official" : "AP",
            "set" : "SetP",
            "recall" : "R@5",
            "recall_1000" : "recall_1000"
        }
        # what we ask for -> what we should NOT get
        family2black = {
            'ndcg_cut_5' : 'ndcg_cut_10',
            'P_5' : "P_100",
            "recall_1000" : "recall_5"
        }
        for m in family2measure:
            df1 = pt.Experiment(res, topics, qrels, eval_metrics=[m])
            df2 = pt.Experiment(res, topics, qrels, eval_metrics=[m], baseline=0)
            df3 = pt.Experiment(res, topics, qrels, eval_metrics=[m], perquery=True)
            self.assertIn(family2measure[m], df1.columns)
            self.assertIn(family2measure[m], df2.columns)
            self.assertTrue(len(df3[df3["measure"] == family2measure[m]])>0)

            # check that we dont get back measures that we did NOT ask for
            if m in family2black:
                self.assertNotIn(family2black[m], df1.columns)
                self.assertNotIn(family2black[m], df2.columns)
                self.assertTrue(len(df3[df3["measure"] == family2black[m]])==0)

    def test_differing_order(self):
        topics = pd.DataFrame([["q1", "q1"], ["q2", "q1"] ], columns=["qid", "query"])
        res1 = pd.DataFrame([ ["q2", "d1", 2.0], ["q1", "d1", 1.0],], columns=["qid", "docno", "score"])
        res2 = pd.DataFrame([["q1", "d1", 1.0], ["q2", "d1", 2.0] ], columns=["qid", "docno", "score"])
        qrels = pd.DataFrame([["q1", "d1", 1], ["q2", "d3", 1] ], columns=["qid", "docno", "label"])
        measures = pt.Experiment(
                [pt.Transformer.from_df(res1, uniform=True), pt.Transformer.from_df(res2, uniform=True)],
                topics,
                qrels,
                ["map"],
                baseline=0)
        self.assertEqual(measures.iloc[0]["map"], 0.5)
        self.assertEqual(measures.iloc[1]["map"], 0.5)
        self.assertEqual(measures.iloc[1]["map +"], 0)
        self.assertEqual(measures.iloc[1]["map -"], 0)

    def test_one_row(self):
        from pyterrier.measures import NumQ
        vaswani = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.Retriever(self._vaswani_index(), wmodel='BM25')
        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg", "num_q"])
        self.assertEqual(10, rtr.iloc[0]["num_q"])

        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), [NumQ])
        self.assertEqual(10, rtr.iloc[0]["NumQ"])
        
    def test_one_row_round(self):
        import pyterrier as pt
        vaswani = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.Retriever(self._vaswani_index(), wmodel='DPH')
        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg", "num_q", pt.measures.NDCG@5], round=2)
        self.assertEqual(str(rtr.iloc[0]["map"]), "0.31")
        self.assertEqual(str(rtr.iloc[0]["nDCG@5"]), "0.46")

        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg", "num_q", pt.measures.NDCG@5], round={"nDCG@5" : 1})
        self.assertEqual(str(rtr.iloc[0]["nDCG@5"]), "0.5")

    def test_batching(self):
        vaswani = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.Retriever(self._vaswani_index(), wmodel='BM25')
        rtr1 = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg", "num_q", "mrt"])
        rtr2 = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg", "num_q", "mrt"], batch_size=2)
        self.assertTrue("mrt" in rtr1.columns)
        self.assertTrue("mrt" in rtr2.columns)
        rtr1.drop(columns=["mrt"], inplace=True)
        rtr2.drop(columns=["mrt"], inplace=True)
        pd.testing.assert_frame_equal(rtr1, rtr2)
        
        rtr1 = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg", "num_q"], perquery=True)
        rtr2 = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg", "num_q"], batch_size=2, perquery=True)
        pd.testing.assert_frame_equal(rtr1, rtr2)

    def test_perquery(self):
        vaswani = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.Retriever(self._vaswani_index(), wmodel='BM25')
        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg"], perquery=True)
        self.assertIsInstance(rtr, pd.DataFrame)

        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg"], perquery='both')
        self.assertIsInstance(rtr, tuple)
        self.assertIn('_repr_html_', dir(rtr))
        self.assertIsInstance(rtr[0], pd.DataFrame)
        self.assertIsInstance(rtr[1], pd.DataFrame)

        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg"], perquery=True, dataframe=False)
        self.assertFalse(isinstance(rtr, pd.DataFrame))

        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg"], perquery='both', dataframe=False)
        self.assertIsInstance(rtr, tuple)
        print(rtr)
        self.assertFalse('_repr_html_' in dir(rtr))
        self.assertFalse(isinstance(rtr[0], pd.DataFrame))
        self.assertFalse(isinstance(rtr[1], pd.DataFrame))

    def test_perquery_round(self):
        vaswani = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.Retriever(self._vaswani_index(), wmodel='DPH')
        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg"], perquery=True, round=2)
        self.assertEqual(str(rtr.iloc[0]["value"]), "0.36")

    def test_bad_measure(self):
        vaswani = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.Retriever(self._vaswani_index(), wmodel='BM25')
        with self.assertRaises(KeyError):
            pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), [map])

    def test_baseline_and_tests(self):
        dataset = pt.get_dataset("vaswani")
        numt=10
        res1 = pt.terrier.Retriever(self._vaswani_index(), wmodel="BM25")(dataset.get_topics().head(numt))
        res2 = pt.terrier.Retriever(self._vaswani_index(), wmodel="DPH")(dataset.get_topics().head(numt))

        # t-test
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            df = pt.Experiment(
                [res1, res2], 
                dataset.get_topics().head(numt), 
                dataset.get_qrels(),
                eval_metrics=["map", "ndcg"], 
                baseline=0)
            self.assertTrue("map +" in df.columns)
            self.assertTrue("map -" in df.columns)
            self.assertTrue("map p-value" in df.columns)

            # wilcoxon signed-rank test
            df = pt.Experiment(
                [res1, res2], 
                dataset.get_topics().head(numt), 
                dataset.get_qrels(),
                eval_metrics=["map", "ndcg"], 
                test='wilcoxon', 
                baseline=0)
            self.assertTrue("map +" in df.columns)
            self.assertTrue("map -" in df.columns)
            self.assertTrue("map p-value" in df.columns)


        # user-specified TOST
        # TOST will omit warnings here, due to low numbers of topics
        import statsmodels.stats.weightstats
        fn = lambda X,Y: (0, statsmodels.stats.weightstats.ttost_paired(X, Y, -0.01, 0.01)[0])
        
        #This filter doesnt work
        with warnings.catch_warnings(record=True) as w:
            warnings.filterwarnings("always")
            df = pt.Experiment(
                [res1, res2], 
                dataset.get_topics().head(numt), 
                dataset.get_qrels(),
                eval_metrics=["map", "ndcg"], 
                test=fn,
                baseline=0)
            print(w)
        self.assertTrue("map +" in df.columns)
        self.assertTrue("map -" in df.columns)
        self.assertTrue("map p-value" in df.columns)
        

    def test_baseline_correction_userdefined_test(self):
        dataset = pt.get_dataset("vaswani")
        res1 = pt.terrier.Retriever(self._vaswani_index(), wmodel="BM25")(dataset.get_topics().head(10))
        res2 = pt.terrier.Retriever(self._vaswani_index(), wmodel="DPH")(dataset.get_topics().head(10))
        # TOST will omit warnings here, due to low numbers of topics
        import statsmodels.stats.weightstats
        fn = lambda X,Y: (0, statsmodels.stats.weightstats.ttost_ind(X, Y, -0.01, 0.01)[0])
        for corr in ['hs', 'bonferroni', 'holm-sidak']:            
            df = pt.Experiment(
                [res1, res2], 
                dataset.get_topics().head(10), 
                dataset.get_qrels(),
                eval_metrics=["map", "ndcg"], 
                baseline=0, correction='hs', test=fn)
            self.assertTrue("map +" in df.columns)
            self.assertTrue("map -" in df.columns)
            self.assertTrue("map p-value" in df.columns)
            self.assertTrue("map p-value corrected" in df.columns)
            self.assertTrue("map reject" in df.columns)

    def test_baseline_corrected(self):
        dataset = pt.get_dataset("vaswani")
        res1 = pt.terrier.Retriever(self._vaswani_index(), wmodel="BM25")(dataset.get_topics().head(10))
        res2 = pt.terrier.Retriever(self._vaswani_index(), wmodel="DPH")(dataset.get_topics().head(10))
        baseline = 0
        for corr in ['hs', 'bonferroni', 'hommel']:
            df = pt.Experiment(
                [res1, res2], 
                dataset.get_topics().head(10), 
                dataset.get_qrels(),
                eval_metrics=["map", "ndcg"], 
                baseline=baseline, correction=corr)
            self.assertTrue("map +" in df.columns)
            self.assertTrue("map -" in df.columns)
            self.assertTrue("map p-value" in df.columns)
            self.assertTrue("map p-value corrected" in df.columns)
            self.assertTrue("map reject" in df.columns)
            self.assertFalse(any(df["map p-value corrected"].drop(df.index[baseline]).isna()))
